"""Scheduling tools based on the schedule module."""

from __future__ import annotations

import logging
from datetime import datetime, timedelta
from typing import TYPE_CHECKING

from schedule import Scheduler

if TYPE_CHECKING:
    from jobflow import Job

logger = logging.getLogger(__name__)


# TODO consider making this with an exponential backoff strategy
# with a failure at the end
class SafeScheduler(Scheduler):
    """
    An implementation of Scheduler that catches jobs that fail, logs their
    exception tracebacks as errors, optionally reschedules the jobs for their
    next run time, and keeps going.

    Adapted from https://gist.github.com/mplewis/8483f1c24f2d6259aef6
    """

    def __init__(
        self, reschedule_on_failure: bool = True, seconds_after_failure: int = 0
    ) -> None:
        """
        If reschedule_on_failure is True, jobs will be rescheduled for their
        next run as if they had completed successfully. If False, they'll run
        on the next run_pending() tick.
        """
        self.reschedule_on_failure = reschedule_on_failure
        self.seconds_after_failure = seconds_after_failure
        super().__init__()

    def _run_job(self, job: Job) -> None:
        try:
            super()._run_job(job)
        except Exception:
            task_name = job.job_func.__name__
            logger.exception(f"Error while running task {task_name}")
            if self.reschedule_on_failure:
                if secs := self.seconds_after_failure:
                    logger.warning(f"Task {task_name} rescheduled in {secs} seconds")
                    job.last_run = None
                    job.next_run = datetime.now() + timedelta(seconds=secs)
                else:
                    logger.warning(f"Task {task_name} rescheduled")
                    job.last_run = datetime.now()
                    job._schedule_next_run()
            else:
                logger.warning(f"Task {task_name} canceled.")
                self.cancel_job(job)



from datetime import datetime
from typing import Annotated, Optional

import click
import typer

from jobflow_remote.cli.utils import (
    IndexDirection,
    SerializeFileFormat,
    SortOption,
    str_to_dict,
    tree_callback,
)
from jobflow_remote.config.base import LogLevel
from jobflow_remote.jobs.state import FlowState, JobState

tree_opt = Annotated[
    bool,
    typer.Option(
        "--tree",
        help="Display a tree representation of the CLI command structure",
        is_eager=True,
        callback=tree_callback,
    ),
]

job_ids_indexes_opt = Annotated[
    Optional[list[str]],
    typer.Option(
        "--job-id",
        "-jid",
        help="One or more pair of job ids (i.e. uuids) and job index formatted "
        "as UUID:INDEX (e.g. e1d66c4f-81db-4fff-bda2-2bf1d79d5961:2). "
        "The index is mandatory",
    ),
]


job_ids_opt = Annotated[
    Optional[list[str]],
    typer.Option(
        "--job-id",
        "-jid",
        help="One or more job ids (i.e. uuids). Only the id is needed since "
        "jobs with the same uuid belong to the same flow",
    ),
]


db_ids_opt = Annotated[
    Optional[list[str]],
    typer.Option(
        "--db-id",
        "-did",
        help="One or more db ids",
    ),
]


flow_ids_opt = Annotated[
    Optional[list[str]],
    typer.Option(
        "--flow-id",
        "-fid",
        help="One or more flow ids",
    ),
]


job_state_opt = Annotated[
    Optional[list[JobState]],
    typer.Option(
        "--state",
        "-s",
        help="One or more of the Job states",
    ),
]


flow_state_opt = Annotated[
    Optional[list[FlowState]],
    typer.Option(
        "--state",
        "-s",
        help="One or more of the Flow states",
    ),
]


name_opt = Annotated[
    Optional[str],
    typer.Option(
        "--name",
        "-n",
        help="The name. Default is an exact match, but all conventions from "
        "python fnmatch can be used (e.g. *test*). Using * wildcard may require"
        "enclosing the search string in quotation marks.",
    ),
]

worker_name_opt = Annotated[
    Optional[list[str]],
    typer.Option(
        "--worker",
        "-wk",
        help="One or more worker names",
    ),
]


job_state_arg = Annotated[JobState, typer.Argument(help="One of the job states")]


start_date_opt = Annotated[
    Optional[datetime],
    typer.Option(
        "--start-date",
        "-sdate",
        help="Initial date for last update field",
        formats=["%Y-%m-%d", "%Y-%m-%dT%H:%M:%S", "%Y-%m-%d", "%H:%M:%S", "%H:%M:%S"],
    ),
]


end_date_opt = Annotated[
    Optional[datetime],
    typer.Option(
        "--end-date",
        "-edate",
        help="Final date for last update field",
    ),
]


days_opt = Annotated[
    Optional[int],
    typer.Option(
        "--days",
        "-ds",
        help="Last update field is in the last days",
    ),
]


hours_opt = Annotated[
    Optional[int],
    typer.Option(
        "--hours",
        "-hs",
        help="Last update field is in the last hours",
    ),
]


verbosity_opt = Annotated[
    int,
    typer.Option(
        "--verbosity",
        "-v",
        help="Set the verbosity of the output. Multiple -v options "
        "increase the verbosity. (e.g. -vvv)",
        count=True,
    ),
]


log_level_opt = Annotated[
    LogLevel,
    typer.Option(
        "--log-level",
        "-log",
        help="Set the log level of the runner",
    ),
]

runner_num_procs_opt = Annotated[
    int,
    typer.Option(
        "--num-procs",
        "-n",
        help="The number of Runner processes started",
    ),
]

max_results_opt = Annotated[
    int,
    typer.Option(
        "--max-results",
        "-m",
        help="Limit the maximum number of returned results. Set 0 for no limit",
    ),
]


sort_opt = Annotated[
    SortOption,
    typer.Option(
        "--sort",
        help="The field on which the results will be sorted. In descending order",
    ),
]


reverse_sort_flag_opt = Annotated[
    bool,
    typer.Option(
        "--reverse-sort",
        "-revs",
        help="Reverse the sorting order",
    ),
]


job_db_id_arg = Annotated[
    str,
    typer.Argument(
        help="The ID of the job. Can be the db id (i.e. an integer) or a string (i.e. the uuid)",
        metavar="ID",
    ),
]
job_index_arg = Annotated[
    Optional[int],
    typer.Argument(
        help="The index of the job. If not defined the job with the largest index is selected",
        metavar="INDEX",
    ),
]

job_index_opt = Annotated[
    Optional[int],
    typer.Option(
        "--index",
        "-i",
        help="The index of the job. If not defined the job with the largest index is selected",
    ),
]


flow_db_id_arg = Annotated[
    str,
    typer.Argument(
        help="The ID of the flow. Can the db id (i.e. an integer) or a string (i.e. the uuid)",
        metavar="ID",
    ),
]


force_opt = Annotated[
    bool,
    typer.Option(
        "--force",
        "-f",
        help="No confirmation will be asked before proceeding",
    ),
]


job_flow_id_flag_opt = Annotated[
    bool,
    typer.Option(
        "--job",
        "-j",
        help="The passed ID will be the ID of one of the jobs"
        " belonging to the flow, instead of the ID of the flow.",
    ),
]

locked_opt = Annotated[
    bool,
    typer.Option(
        "--locked",
        "-l",
        help="Select locked Jobs",
    ),
]


locked_flow_opt = Annotated[
    bool,
    typer.Option(
        "--locked",
        "-l",
        help="Select locked Flows",
    ),
]


serialize_file_format_opt = Annotated[
    SerializeFileFormat,
    typer.Option(
        "--format",
        "-f",
        help="File format",
    ),
]


wait_lock_opt = Annotated[
    int,
    typer.Option(
        "--wait",
        "-w",
        help="When trying to acquire the lock on the documents that need to "
        "be modified, wait an amount of seconds equal to the value specified",
    ),
]


break_lock_opt = Annotated[
    bool,
    typer.Option(
        "--break-lock",
        "-bl",
        help="Forcibly break the lock for the documents that need to be modified. "
        "Use with care and possibly when the runner is stopped. Can lead to "
        "inconsistencies",
    ),
]


raise_on_error_opt = Annotated[
    bool,
    typer.Option(
        "--raise-on-error",
        "-re",
        help="If an error arises during any of the operations raise an exception "
        "and stop the execution",
    ),
]


delete_output_opt = Annotated[
    bool,
    typer.Option(
        "--output",
        "-o",
        help="Also delete the outputs of the Jobs in the output Store",
    ),
]


delete_files_opt = Annotated[
    bool,
    typer.Option(
        "--files",
        "-fi",
        help="Also delete the files on the worker",
    ),
]


delete_all_opt = Annotated[
    bool,
    typer.Option(
        "--all",
        "-a",
        help="enable --output and --files",
    ),
]

foreground_index_opt = Annotated[
    bool,
    typer.Option(
        "--foreground",
        "-fg",
        help="The build of the indexes will not be executed in the background",
    ),
]

index_key_arg = Annotated[
    str,
    typer.Argument(
        help="The field on which the index will be created",
        metavar="INDEX",
    ),
]

index_direction_arg = Annotated[
    Optional[IndexDirection],
    typer.Argument(
        help="The direction of the index",
        metavar="DIRECTION",
    ),
]


# as of typer version 0.9.0 the dict is not a supported type. Define a custom one
class DictType(dict):
    pass


# Similarly, Python 3.10 union types are not supported,
# e.g., `str | None` vs `Optional[str]`
# ruff enforces PEP 604
# but will leave things in if they are explicit type aliases
OptionalStr = Optional[str]
OptionalDictType = Optional[DictType]


class DictTypeParser(click.ParamType):
    name = "DictType"

    def convert(self, value, param, ctx):
        value = str_to_dict(value)
        return DictType(value)


query_opt = Annotated[
    OptionalDictType,
    typer.Option(
        "--query",
        "-q",
        help="A JSON string representing a generic query in the form of a dictionary. "
        "Overrides all other query options. Requires knowledge of the internal structure of the DB. "
        "Can be either a list of comma separated key=value pairs or a string with the JSON"
        " representation of a dictionary containing the mongoDB query that "
        'should be performed (e.g \'{"key1.key2": 1, "key3": "test"}\')',
        click_type=DictTypeParser(),
    ),
]


metadata_opt = Annotated[
    OptionalDictType,
    typer.Option(
        "--metadata",
        "-meta",
        help="A string representing the metadata to be queried. Can be either"
        " a list of comma separated key=value pairs or a string with the JSON"
        " representation of a dictionary containing the mongoDB query for "
        'the metadata subdocument (e.g \'{"key1.key2": 1, "key3": "test"}\')',
        click_type=DictTypeParser(),
    ),
]



"""jobflow-remote is a python package to run jobflow workflows on remote resources."""

from jobflow_remote._version import __version__
from jobflow_remote.config.jobconfig import set_run_config
from jobflow_remote.config.manager import ConfigManager
from jobflow_remote.config.settings import JobflowRemoteSettings
from jobflow_remote.jobs.jobcontroller import JobController
from jobflow_remote.jobs.store import get_jobstore
from jobflow_remote.jobs.submit import submit_flow

SETTINGS = JobflowRemoteSettings()

__all__ = (
    "__version__",
    "set_run_config",
    "ConfigManager",
    "JobflowRemoteSettings",
    "JobController",
    "get_jobstore",
    "submit_flow",
)



from pathlib import Path
from typing import Optional

from pydantic import Field, model_validator
from pydantic_settings import BaseSettings, SettingsConfigDict

from jobflow_remote.config.base import LogLevel

DEFAULT_PROJECTS_FOLDER = Path("~/.jfremote").expanduser().as_posix()

DEFAULT_CONFIG_FILE_PATH = Path("~/.jfremote.yaml").expanduser().as_posix()


class JobflowRemoteSettings(BaseSettings):
    config_file: str = Field(
        DEFAULT_CONFIG_FILE_PATH,
        description="Location of the config file for jobflow remote.",
    )
    projects_folder: str = Field(
        DEFAULT_PROJECTS_FOLDER, description="Location of the projects files."
    )
    project: Optional[str] = Field(None, description="The name of the project used.")
    cli_full_exc: bool = Field(
        default=False,
        description="If True prints the full stack trace of the exception when raised in the CLI.",
    )
    cli_suggestions: bool = Field(
        default=True, description="If True prints some suggestions in the CLI commands."
    )
    cli_log_level: LogLevel = Field(
        LogLevel.WARN, description="The level set for logging in the CLI"
    )
    cli_job_list_columns: Optional[list[str]] = Field(
        None,
        description="The list of columns to show in the `jf job list` command. For available "
        "options check the corresponding help: `jf job list -h`.",
    )

    model_config = SettingsConfigDict(env_prefix="jfremote_")

    @model_validator(mode="before")
    @classmethod
    def load_default_settings(cls, values: dict) -> dict:
        """
        Load settings from file or environment variables.

        Loads settings from a root file if available and uses that as defaults in
        place of built-in defaults.

        This allows setting of the config file path through environment variables.
        """
        from monty.serialization import loadfn

        config_file_path: str = values.get("config_file", DEFAULT_CONFIG_FILE_PATH)

        new_values = {}
        if Path(config_file_path).expanduser().exists():
            new_values.update(loadfn(Path(config_file_path).expanduser()))

        new_values.update(values)
        return new_values



from typing import Callable

import typer
from typer.models import CommandFunctionType

from jobflow_remote.cli.types import tree_opt
from jobflow_remote.cli.utils import cli_error_handler


class JFRTyper(typer.Typer):
    """Subclassing typer to intercept exceptions and print nicer error messages."""

    def __init__(self, *args, **kwargs) -> None:
        if "epilog" not in kwargs:
            kwargs["epilog"] = (
                "Run [bold]'jf -h'[/] to display the [bold]global options[/]"
            )

        if "rich_markup_mode" not in kwargs:
            kwargs["rich_markup_mode"] = "rich"

        if "callback" not in kwargs:
            kwargs["callback"] = default_callback

        # if "result_callback" not in kwargs:
        #     kwargs["result_callback"] = test_cb

        super().__init__(*args, **kwargs)

    def command(
        self, *args, **kwargs
    ) -> Callable[[CommandFunctionType], CommandFunctionType]:
        typer_wrapper = super().command(*args, **kwargs)

        def wrapper(fn):
            fn = cli_error_handler(fn)
            return typer_wrapper(fn)

        return wrapper

    def callback(
        self, *args, **kwargs
    ) -> Callable[[CommandFunctionType], CommandFunctionType]:
        typer_wrapper = super().callback(*args, **kwargs)

        def wrapper(fn):
            fn = cli_error_handler(fn)
            return typer_wrapper(fn)

        return wrapper


def default_callback(print_tree: tree_opt = False):
    pass



from __future__ import annotations

import datetime
import glob
import logging
import os
import subprocess
import time
import traceback
from multiprocessing import Manager, Process
from typing import TYPE_CHECKING

from jobflow import JobStore, initialize_logger
from jobflow.core.flow import get_flow
from monty.design_patterns import singleton
from monty.os import cd
from monty.serialization import dumpfn, loadfn
from monty.shutil import decompress_file

from jobflow_remote.jobs.batch import LocalBatchManager
from jobflow_remote.jobs.data import IN_FILENAME, OUT_FILENAME, JobDoc
from jobflow_remote.remote.data import get_job_path, get_store_file_paths
from jobflow_remote.utils.log import initialize_remote_run_log

if TYPE_CHECKING:
    from pathlib import Path

    from jobflow.core.job import Job

logger = logging.getLogger(__name__)


@singleton
class JfrState:
    """State of the current job being executed."""

    job_doc: JobDoc = None

    def reset(self):
        """Reset the current state."""
        self.job_doc = None


CURRENT_JOBDOC: JfrState = JfrState()


def run_remote_job(run_dir: str | Path = ".") -> None:
    """Run the job."""
    initialize_remote_run_log()

    start_time = datetime.datetime.utcnow()
    with cd(run_dir):
        error = None
        try:
            dumpfn({"start_time": start_time}, OUT_FILENAME)
            in_data = loadfn(IN_FILENAME)

            job: Job = in_data["job"]
            store = in_data["store"]
            job_doc_dict = in_data.get("job_doc", None)
            if job_doc_dict:
                job_doc_dict["job"] = job
                JfrState().job_doc = JobDoc.model_validate(job_doc_dict)

            store.connect()

            initialize_logger()
            try:
                response = job.run(store=store)
            finally:
                # some jobs may have compressed the jfremote and store files while being
                # executed, try to decompress them if that is the case and files need to be
                # decompressed.
                decompress_files(store)

            # Close the store explicitly, as minimal stores may require it.
            try:
                store.close()
            except Exception:
                logger.exception("Error while closing the store")

            # The output of the response has already been stored in the store.
            response.output = None

            # Convert to Flow the dynamic responses before dumping the output.
            # This is required so that the response does not need to be
            # deserialized and converted to Flows by the runner.
            if response.addition:
                response.addition = get_flow(response.addition)
            if response.detour:
                response.detour = get_flow(response.detour)
            if response.replace:
                response.replace = get_flow(response.replace)

            output = {
                "response": response,
                "error": error,
                "start_time": start_time,
                "end_time": datetime.datetime.utcnow(),
            }
            dumpfn(output, OUT_FILENAME)
        except Exception:
            # replicate the dump to catch potential errors in
            # serializing/dumping the response.
            error = traceback.format_exc()
            output = {
                "response": None,
                "error": error,
                "start_time": start_time,
                "end_time": datetime.datetime.utcnow(),
            }
            dumpfn(output, OUT_FILENAME)
        finally:
            JfrState().reset()


def run_batch_jobs(
    base_run_dir: str | Path,
    files_dir: str | Path,
    process_uuid: str,
    max_time: int | None = None,
    max_wait: int = 60,
    max_jobs: int | None = None,
    parallel_jobs: int | None = None,
) -> None:
    parallel_jobs = parallel_jobs or 1

    if parallel_jobs == 1:
        run_single_batch_jobs(
            base_run_dir=base_run_dir,
            files_dir=files_dir,
            process_uuid=process_uuid,
            max_time=max_time,
            max_wait=max_wait,
            max_jobs=max_jobs,
        )
    else:
        with Manager() as manager:
            multiprocess_lock = manager.Lock()
            parallel_ids = manager.dict()
            batch_manager = LocalBatchManager(
                files_dir=files_dir,
                process_id=process_uuid,
                multiprocess_lock=multiprocess_lock,
            )
            processes = [
                Process(
                    target=run_single_batch_jobs,
                    args=(
                        base_run_dir,
                        files_dir,
                        process_uuid,
                        max_time,
                        max_wait,
                        max_jobs,
                        batch_manager,
                        parallel_ids,
                    ),
                )
                for _ in range(parallel_jobs)
            ]
            for p in processes:
                p.start()
                time.sleep(0.5)

            for p in processes:
                p.join()


def run_single_batch_jobs(
    base_run_dir: str | Path,
    files_dir: str | Path,
    process_uuid: str,
    max_time: int | None = None,
    max_wait: int = 60,
    max_jobs: int | None = None,
    batch_manager: LocalBatchManager | None = None,
    parallel_ids: dict | None = None,
) -> None:
    initialize_remote_run_log()

    # TODO the ID should be somehow linked to the queue job
    if not batch_manager:
        batch_manager = LocalBatchManager(files_dir=files_dir, process_id=process_uuid)

    if parallel_ids:
        parallel_ids[os.getpid()] = False

    t0 = time.time()
    wait = 0
    sleep_time = 10
    count = 0
    while True:
        if max_time and max_time < time.time() - t0:
            logger.info("Stopping due to max_time")
            return

        if max_wait and wait > max_wait:
            # if many jobs run in parallel do not shut down here, unless all
            # the other jobs are also stopped
            if parallel_ids:
                for pid, pid_is_running in parallel_ids.items():
                    if pid_is_running:
                        try:
                            os.kill(pid, 0)  # throws OSError if the process is dead
                        except OSError:  # means this process is dead!
                            pid_is_running[pid] = False
                if not any(parallel_ids.values()):
                    logger.info(
                        f"No jobs available for more than {max_wait} seconds and all other jobs are stopped. Stopping."
                    )
                    return
            else:
                logger.info(
                    f"No jobs available for more than {max_wait} seconds. Stopping."
                )
                return

        if max_jobs and count >= max_jobs:
            logger.info(f"Maximum number of jobs reached ({max_jobs}). Stopping.")
            return

        job_str = batch_manager.get_job()
        if not job_str:
            time.sleep(sleep_time)
            wait += sleep_time
        else:
            wait = 0
            count += 1
            job_id, _index = job_str.split("_")
            index: int = int(_index)
            logger.info(f"Starting job with id {job_id} and index {index}")
            job_path = get_job_path(job_id=job_id, index=index, base_path=base_run_dir)
            if parallel_ids:
                parallel_ids[os.getpid()] = True
            try:
                with cd(job_path):
                    result = subprocess.run(
                        ["bash", "submit.sh"],  # noqa: S603, S607
                        check=True,
                        text=True,
                        capture_output=True,
                    )
                    if result.returncode:
                        logger.warning(
                            f"Process for job with id {job_id} and index {index} finished with an error"
                        )
                batch_manager.terminate_job(job_id, index)
            except Exception:
                logger.exception(
                    "Error while running job with id {job_id} and index {index}"
                )
            else:
                logger.info(f"Completed job with id {job_id} and index {index}")

        if parallel_ids:
            parallel_ids[os.getpid()] = False


def decompress_files(store: JobStore) -> None:
    file_names = [OUT_FILENAME]
    file_names.extend(os.path.basename(p) for p in get_store_file_paths(store))

    for fn in file_names:
        # If the file is already present do not decompress it, even if
        # a compressed version is present.
        if os.path.isfile(fn):
            continue
        for f in glob.glob(fn + ".*"):
            decompress_file(f)



from __future__ import annotations

import contextlib
import datetime
import getpass
import logging
import os
import re
import socket
import subprocess
import time
from enum import Enum
from pathlib import Path
from string import Template
from typing import TYPE_CHECKING, Callable
from xmlrpc.client import Fault

import psutil
from monty.os import makedirs_p
from supervisor import childutils, states, xmlrpc
from supervisor.compat import xmlrpclib
from supervisor.options import ClientOptions
from supervisor.states import RUNNING_STATES, STOPPED_STATES, ProcessStates
from supervisor.supervisorctl import Controller, fgthread
from supervisor.xmlrpc import Faults

from jobflow_remote.config import ConfigManager, Project
from jobflow_remote.jobs.jobcontroller import JobController
from jobflow_remote.utils.db import MongoLock, RunnerLockedError

if TYPE_CHECKING:
    from collections.abc import Generator

logger = logging.getLogger(__name__)


supervisord_conf_str = """
[unix_http_server]
file=$sock_file

[rpcinterface:supervisor]
supervisor.rpcinterface_factory = supervisor.rpcinterface:make_main_rpcinterface

[supervisord]
logfile=$log_file
logfile_maxbytes=10MB
logfile_backups=5
loglevel=$loglevel
pidfile=$pid_file
nodaemon=$nodaemon

[supervisorctl]
serverurl=unix://$sock_file

[program:runner_daemon]
priority=100
command=jf -p $project runner run -pid -log $loglevel $connect_interactive
autostart=true
autorestart=false
numprocs=1
process_name=run_jobflow%(process_num)s
stopwaitsecs=86400
"""


supervisord_conf_str_split = """
[unix_http_server]
file=$sock_file

[rpcinterface:supervisor]
supervisor.rpcinterface_factory = supervisor.rpcinterface:make_main_rpcinterface

[supervisord]
logfile=$log_file
logfile_maxbytes=10MB
logfile_backups=5
loglevel=$loglevel
pidfile=$pid_file
nodaemon=$nodaemon

[supervisorctl]
serverurl=unix://$sock_file

[program:runner_daemon_checkout]
priority=100
command=jf -p $project runner run -pid --checkout -log $loglevel $connect_interactive
autostart=true
autorestart=false
numprocs=1
process_name=run_jobflow_checkout
stopwaitsecs=86400

[program:runner_daemon_transfer]
priority=100
command=jf -p $project runner run -pid --transfer -log $loglevel $connect_interactive
autostart=true
autorestart=false
numprocs=$num_procs_transfer
process_name=run_jobflow_transfer%(process_num)s
stopwaitsecs=86400

[program:runner_daemon_queue]
priority=100
command=jf -p $project runner run -pid --queue -log $loglevel $connect_interactive
autostart=true
autorestart=false
numprocs=1
process_name=run_jobflow_queue
stopwaitsecs=86400

[program:runner_daemon_complete]
priority=100
command=jf -p $project runner run -pid --complete -log $loglevel $connect_interactive
autostart=true
autorestart=false
numprocs=$num_procs_complete
process_name=run_jobflow_complete%(process_num)s
stopwaitsecs=86400
"""


class DaemonError(Exception):
    pass


class RunningDaemonError(Exception):
    pass


class DaemonStatus(Enum):
    """
    Possible states of the daemon.
    """

    SHUT_DOWN = "SHUT_DOWN"
    STOPPED = "STOPPED"
    STOPPING = "STOPPING"
    PARTIALLY_RUNNING = "PARTIALLY_RUNNING"
    STARTING = "STARTING"
    RUNNING = "RUNNING"


class DaemonManager:
    """
    A manager for handling the daemonized Runner processes.

    Checks are performed to avoid starting a daemon in two different machines.
    To perform the actions on the daemon process the information about the runner
    in the DB should match with the one of the current machine.
    """

    conf_template_single = Template(supervisord_conf_str)
    conf_template_split = Template(supervisord_conf_str_split)

    def __init__(
        self,
        daemon_dir: str | Path,
        log_dir: str | Path,
        project: Project,
        job_controller: JobController | None = None,
    ) -> None:
        """
        Parameters
        ----------
        daemon_dir
            Directory where the supervisord configuration file is written.
        log_dir
            Directory where the supervisord log file is written.
        project
            Project configuration.
        job_controller
            JobController instance, used internally to handle the job queue.
        """
        self.project = project
        self.daemon_dir = Path(daemon_dir).absolute()
        self.log_dir = Path(log_dir).absolute()
        self._job_controller = job_controller

    @classmethod
    def from_project(cls, project: Project) -> DaemonManager:
        """
        Generate a DaemonManager instance from a Project configuration.

        Parameters
        ----------
        project
            Project configuration.

        Returns
        -------
        DaemonManager
            An instance of DaemonManager associated with the project.
        """
        daemon_dir = project.daemon_dir
        log_dir = project.log_dir
        return cls(daemon_dir, log_dir, project)

    @classmethod
    def from_project_name(cls, project_name: str | None = None) -> DaemonManager:
        """
        Generate a DaemonManager instance from a project name.

        Parameters
        ----------
        project_name
            Name of the project to use. If None, the default project will be used.

        Returns
        -------
        DaemonManager
            An instance of DaemonManager associated with the project.
        """
        config_manager = ConfigManager()
        project = config_manager.get_project(project_name)
        return cls.from_project(project)

    @property
    def conf_filepath(self) -> Path:
        """
        Path to the supervisord configuration file.
        """
        return self.daemon_dir / "supervisord.conf"

    @property
    def pid_filepath(self) -> Path:
        """
        Path to the supervisord PID file.
        """
        return self.daemon_dir / "supervisord.pid"

    @property
    def log_filepath(self) -> Path:
        """
        Path to the supervisord log file.
        """
        return self.log_dir / "supervisord.log"

    @property
    def sock_filepath(self) -> Path:
        """
        Path to the supervisord socket file.

        This path is used by supervisord to communicate with the client.
        The length of the path is checked to ensure it is not too long for UNIX systems.
        """
        path = self.daemon_dir / "s.sock"
        if len(str(path)) > 97:
            msg = f"socket path {path} is too long for UNIX systems. Set the daemon_dir value "
            "in the project configuration so that the socket path is shorter"
            raise DaemonError(msg)
        return path

    @property
    def job_controller(self) -> JobController:
        """
        JobController instance associated with the project.
        """
        if self._job_controller is None:
            self._job_controller = JobController.from_project(self.project)
        return self._job_controller

    def clean_files(self) -> None:
        """
        Clean up the supervisord PID and socket files.
        """
        self.pid_filepath.unlink(missing_ok=True)
        self.sock_filepath.unlink(missing_ok=True)

    def get_interface(self):
        """
        Return an interface to the supervisord RPC server using the socket file.

        Returns
        -------
        xmlrpc.ServerProxy
            Interface to the supervisord RPC server.
        """
        env = {
            "SUPERVISOR_SERVER_URL": f"unix://{self.sock_filepath!s}",
            "SUPERVISOR_USERNAME": "",
            "SUPERVISOR_PASSWORD": "",
        }
        return childutils.getRPCInterface(env)

    def get_supervisord_pid(self) -> int | None:
        """
        Get the PID of the supervisord process from the PID file.

        If the PID file does not exist or can not be parsed, return None.

        Returns
        -------
        int | None
            PID of the supervisord process.
        """
        pid_fp = self.pid_filepath

        if not pid_fp.is_file():
            return None

        try:
            with open(pid_fp) as f:
                pid = int(f.read().strip())
        except ValueError:
            logger.warning(f"The pid file {pid_fp} could not be parsed")
            return None
        return pid

    def check_supervisord_process(self) -> bool:
        """
        Check if the supervisord process is running and belongs to the same user.

        If the supervisord process is not running but the daemon files are present, clean them up.

        Returns
        -------
        bool
            True if the supervisord process is running and belongs to the same user, False otherwise.
        """
        pid = self.get_supervisord_pid()

        running = True
        if pid is None:
            running = False

        try:
            process = psutil.Process(pid)

            for cmdline_element in process.cmdline():
                if cmdline_element.endswith("supervisord"):
                    break
            else:
                running = False

            if process.username() != psutil.Process().username():
                logger.warning(
                    f"pid {pid} is running supervisord, but belongs to a different user"
                )
                running = False
        except psutil.NoSuchProcess:
            running = False

        if not running and pid is not None:
            logger.warning(
                f"Process with pid {pid} is not running but daemon files are present. Cleaning them up."
            )
            self.clean_files()

        return running

    def check_status(self) -> DaemonStatus:
        """
        Get the current status of the daemon based on the state of the supervisord
        process and the running processes.

        Returns
        -------
        DaemonStatus
            Status of the daemon. Can be one of the following:
        """
        process_active = self.check_supervisord_process()

        if not process_active:
            return DaemonStatus.SHUT_DOWN

        if not self.sock_filepath.is_socket():
            raise DaemonError(
                "the supervisord process is alive, but the socket is missing"
            )

        interface = self.get_interface()
        try:
            proc_info = interface.supervisor.getAllProcessInfo()
        except Fault as exc:
            # catch this exception as it may be raised if the status is queried while
            # the supervisord process is shutting down. The error is quite cryptic, so
            # replace with one that is clearer. Also see a related issue in supervisord:
            # https://github.com/Supervisor/supervisor/issues/48
            if exc.faultString == "SHUTDOWN_STATE":
                raise DaemonError(
                    "The daemon is likely shutting down and the actual state cannot be determined"
                ) from exc
            raise
        if not proc_info:
            raise DaemonError(
                "supervisord process is running but no daemon process is present"
            )

        if all(pi.get("state") in RUNNING_STATES for pi in proc_info):
            if any(pi.get("state") == ProcessStates.STARTING for pi in proc_info):
                return DaemonStatus.STARTING
            return DaemonStatus.RUNNING

        if any(pi.get("state") in RUNNING_STATES for pi in proc_info):
            return DaemonStatus.PARTIALLY_RUNNING

        if all(pi.get("state") in STOPPED_STATES for pi in proc_info):
            return DaemonStatus.STOPPED

        if all(
            pi.get("state") in (ProcessStates.STOPPED, ProcessStates.STOPPING)
            for pi in proc_info
        ):
            return DaemonStatus.STOPPING

        raise DaemonError("Could not determine the current status of the daemon")

    def get_processes_info(self) -> dict[str, dict] | None:
        """
        Get the information about the processes of the daemon.

        None if the daemon is not running.

        Returns
        -------
        dict
            A dictionary with the information about the processes.
            The keys are the process names and the values are dictionaries
            with the process information.
        """
        process_active = self.check_supervisord_process()

        if not process_active:
            return None

        pids = {
            "supervisord": {
                "pid": self.get_supervisord_pid(),
                "statename": "RUNNING",
                "state": ProcessStates.RUNNING,
                "group": None,
            }
        }

        if not self.sock_filepath.is_socket():
            raise DaemonError(
                "the supervisord process is alive, but the socket is missing"
            )

        interface = self.get_interface()
        proc_info = interface.supervisor.getAllProcessInfo()
        if not proc_info:
            raise DaemonError(
                "supervisord process is running but no daemon process is present"
            )

        for pi in proc_info:
            name = f"{pi.get('group')}:{pi.get('name')}"
            pids[name] = pi

        return pids

    def write_config(
        self,
        num_procs_transfer: int = 1,
        num_procs_complete: int = 1,
        single: bool = True,
        log_level: str = "info",
        nodaemon: bool = False,
        connect_interactive: bool = False,
    ) -> None:
        """
        Write the configuration file for the daemon.

        If `single` is True, a configuration to execute all the Runner tasks
        in a single process is written. Otherwise, a split configuration is
        written with separate sections for each task of the Runner.

        Parameters
        ----------
        num_procs_transfer : int
            Number of processes to use for the transfer step.
        num_procs_complete : int
            Number of processes to use for the complete step.
        single : bool, optional
            Write a single configuration file instead of a split one.
        log_level : str, optional
            The log level to use for the daemon.
        nodaemon : bool, optional
            Run the daemon in foreground.
        connect_interactive : bool, optional
            Allow the daemon to perform an interactive initial setup.
        """
        if single:
            conf = self.conf_template_single.substitute(
                sock_file=str(self.sock_filepath),
                pid_file=str(self.pid_filepath),
                log_file=str(self.log_filepath),
                nodaemon="true" if nodaemon else "false",
                project=self.project.name,
                loglevel=log_level,
                connect_interactive=(
                    "--connect-interactive" if connect_interactive else ""
                ),
            )
        else:
            conf = self.conf_template_split.substitute(
                sock_file=str(self.sock_filepath),
                pid_file=str(self.pid_filepath),
                log_file=str(self.log_filepath),
                num_procs_transfer=num_procs_transfer,
                num_procs_complete=num_procs_complete,
                nodaemon="true" if nodaemon else "false",
                project=self.project.name,
                loglevel=log_level,
                connect_interactive=(
                    "--connect-interactive" if connect_interactive else ""
                ),
            )
        with open(self.conf_filepath, "w") as f:
            f.write(conf)

    def start_supervisord(
        self,
        num_procs_transfer: int = 1,
        num_procs_complete: int = 1,
        single: bool = True,
        log_level: str = "info",
        nodaemon: bool = False,
        connect_interactive: bool = False,
    ) -> str | None:
        """
        Start the supervisord daemon and all the processes.

        Parameters
        ----------
        num_procs_transfer
            Number of processes to use for the transfer step.
        num_procs_complete
            Number of processes to use for the complete step.
        single
            Write a single configuration file instead of a split one.
        log_level
            The log level to use for the daemon.
        nodaemon : bool, optional
            Run the daemon in foreground.
        connect_interactive : bool, optional
            Allow the daemon to perform an interactive initial setup.

        Returns
        -------
        str | None
            An error message if the daemon could not be started, otherwise None.
        """
        makedirs_p(self.daemon_dir)
        makedirs_p(self.log_dir)
        self.write_config(
            num_procs_transfer=num_procs_transfer,
            num_procs_complete=num_procs_complete,
            single=single,
            log_level=log_level,
            nodaemon=nodaemon,
            connect_interactive=connect_interactive,
        )
        cp = subprocess.run(
            f"supervisord -c {self.conf_filepath!s}",
            shell=True,
            capture_output=True,
            text=True,
            check=False,
        )
        if cp.returncode != 0:
            return f"Error staring the supervisord process. stdout: {cp.stdout}. stderr: {cp.stderr}"

        # TODO check if actually started?

        return None

    def start_processes(self) -> str | None:
        """
        Start all the processes of the daemon when the supervisord processes
        is running but the daemon processes are stopped.

        Returns
        -------
        str | None
            An error message if not all the processes started correctly, otherwise None.
        """
        interface = self.get_interface()
        result = interface.supervisor.startAllProcesses()
        if not result:
            return "No process started"

        failed = [r for r in result if r.get("status") != Faults.SUCCESS]
        if len(failed) == 0:
            return None

        msg = "Not all the daemon processes started correctly. Details: \n"
        for f in failed:
            msg += f"  - {f.get('description')} - {f.get('status')}\n"
        return msg

    def start(
        self,
        num_procs_transfer: int = 1,
        num_procs_complete: int = 1,
        single: bool = True,
        log_level: str = "info",
        raise_on_error: bool = False,
        connect_interactive: bool = False,
    ) -> bool:
        """
        Start the daemon by starting the supervisord process and all the processes of the daemon.

        Parameters
        ----------
        num_procs_transfer
            Number of processes to use for the transfer of jobs.
        num_procs_complete
            Number of processes to use for the completion of jobs.
        single
            If True, the runner will be started with a single process.
        log_level : str
            Log level of the daemon.
        raise_on_error : bool
            If True, raise an exception if an error occurs.
        connect_interactive : bool
            Allow the daemon to perform an interactive initial setup.

        Returns
        -------
        bool
            True if the daemon is started correctly, False otherwise.
        """
        with self.lock_runner_doc() as lock:
            doc = lock.locked_document
            doc_error = self._check_running_runner(doc, raise_on_error=raise_on_error)
            if doc_error:
                logger.error(doc_error)
                return False
            status = self.check_status()
            if status == DaemonStatus.RUNNING:
                error = "Daemon process is already running"
            elif status == DaemonStatus.SHUT_DOWN:
                error = self.start_supervisord(
                    num_procs_transfer=num_procs_transfer,
                    num_procs_complete=num_procs_complete,
                    single=single,
                    log_level=log_level,
                    connect_interactive=connect_interactive,
                )
            elif status == DaemonStatus.STOPPED:
                # Supervisor sets the processes configuration when the supervisord
                # process is started. In the STOPPED state it is not possible to switch
                # them. Initially the code used to perform a shut_down here before
                # restarting, but this would require handling the lock mechanism in
                # shut_down() and made pointless the concept of STOP.
                # Now the user is warned that the runner restarted with the same
                # old configuration.
                old_config = self.parse_config_file()
                new_config = {
                    "num_procs_transfer": num_procs_transfer,
                    "num_procs_complete": num_procs_complete,
                    "single": single,
                    "log_level": log_level,
                    "connect_interactive": connect_interactive,
                }
                diffs_config = []
                for config_name, config_value in new_config.items():
                    if config_value != old_config.get(config_name):
                        diffs_config.append(config_name)
                if diffs_config:
                    logger.warning(
                        f"Daemon is {DaemonStatus.STOPPED.value}, but the options {', '.join(diffs_config)} "
                        "differ from the values used to activate supervisor. The daemon will start with the initial "
                        "configurations. To change the configuration shut down the daemon and start it again."
                    )
                error = self.start_processes()
            elif status == DaemonStatus.STOPPING:
                error = "Daemon process are stopping. Cannot start."
            else:
                error = f"Daemon status {status} could not be handled"

            if error is not None:
                if raise_on_error:
                    raise DaemonError(error)
                logger.error(error)
                return False
            lock.update_on_release = {
                "$set": {"running_runner": self._get_runner_info()}
            }
            return True

    def stop(self, wait: bool = False, raise_on_error: bool = False) -> bool:
        """
        Stop the daemon.

        Parameters
        ----------
        wait
            If True wait until the daemon has stopped.
        raise_on_error
            Raise an exception if the daemon cannot be stopped.

        Returns
        -------
        bool
            True if the daemon was stopped successfully, False otherwise.
        """
        with self.lock_runner_doc(allow_missing=True) as lock:
            # This check is done to allow users to switch off the runner
            # even if the running_runner document is not present.
            # If the document is locked the error will be raised by lock_runner_doc
            if lock.locked_document:
                doc_error = self._check_running_runner(
                    lock.locked_document, raise_on_error=raise_on_error
                )
                if doc_error:
                    logger.error(doc_error)
                    return False
            status = self.check_status()
            if status in (
                DaemonStatus.STOPPED,
                DaemonStatus.STOPPING,
                DaemonStatus.SHUT_DOWN,
            ):
                lock.update_on_release = {"$set": {"running_runner": None}}
                return True

            if status in (DaemonStatus.RUNNING, DaemonStatus.PARTIALLY_RUNNING):
                interface = self.get_interface()
                if wait:
                    result = interface.supervisor.stopAllProcesses()
                else:
                    result = interface.supervisor.signalAllProcesses(15)

                error = self._verify_call_result(result, "stop", raise_on_error)

                if error is None:
                    lock.update_on_release = {"$set": {"running_runner": None}}
                return error is None

            raise DaemonError(f"Daemon status {status} could not be handled")

    def _verify_call_result(
        self, result, action: str, raise_on_error: bool = False
    ) -> str | None:
        """
        Verify that the result of the XML-RPC call to send a signal to the
        processes completed correctly.

        Parameters
        ----------
        result
            The result of the XML-RPC call.
        action
            The action that was requested.
        raise_on_error
            If True, raise an exception if the result is not correct.

        Returns
        -------
        str | None
            The error message if the operation was not executed correctly, None otherwise.
        """
        error = None
        if not result:
            error = f"The action {action} was not applied to the processes"
        else:
            failed = [r for r in result if r.get("status") == Faults.SUCCESS]
            if len(failed) != len(result):
                error = f"The action {action} was not applied to all the processes. Details: \n"
                for f in failed:
                    error += f"  - {f.get('description')}\n"

        if error is not None:
            if raise_on_error:
                raise DaemonError(error)
            logger.error(error)
            return error

        return None

    def kill(self, raise_on_error: bool = False) -> bool:
        with self.lock_runner_doc() as lock:
            doc_error = self._check_running_runner(
                lock.locked_document, raise_on_error=raise_on_error
            )
            if doc_error:
                logger.error(doc_error)
                return False
            # If the daemon is shutting down supervisord may not be able to identify
            # the state. Try proceeding in that case, since we really want to kill
            # the process
            status = None
            try:
                status = self.check_status()
                if status == DaemonStatus.SHUT_DOWN:
                    logger.info("supervisord is not running. No process is running")
                    lock.update_on_release = {"$set": {"running_runner": None}}
                    return True
                if status == DaemonStatus.STOPPED:
                    logger.info("Processes are already stopped.")
                    lock.update_on_release = {"$set": {"running_runner": None}}
                    return True
            except DaemonError as e:
                msg = (
                    f"Error while determining the state of the runner: {getattr(e, 'message', str(e))}."
                    f"Proceeding with the kill command."
                )
                logger.warning(msg)

            if status in (
                None,
                DaemonStatus.RUNNING,
                DaemonStatus.STOPPING,
                DaemonStatus.PARTIALLY_RUNNING,
            ):
                interface = self.get_interface()
                result = interface.supervisor.signalAllProcesses(9)
                error = self._verify_call_result(result, "kill", raise_on_error)

                if error is None:
                    lock.update_on_release = {"$set": {"running_runner": None}}
                return error is None

        raise DaemonError(f"Daemon status {status} could not be handled")

    def shut_down(self, raise_on_error: bool = False) -> bool:
        """
        Shut down the supervisord process and all the processes of the daemon.

        Parameters
        ----------
        raise_on_error
            If True, raise an exception if an error occurs.

        Returns
        -------
        bool
            True if the daemon is shut down correctly, False otherwise.
        """
        with self.lock_runner_doc(allow_missing=True) as lock:
            # This check is done to allow users to switch off the runner
            # even if the running_runner document is not present.
            # If the document is locked the error will be raised by lock_runner_doc
            if lock.locked_document:
                doc_error = self._check_running_runner(
                    lock.locked_document, raise_on_error=raise_on_error
                )
                if doc_error:
                    logger.error(doc_error)
                    return False
            status = self.check_status()
            if status == DaemonStatus.SHUT_DOWN:
                logger.info("supervisord is already shut down.")
                lock.update_on_release = {"$set": {"running_runner": None}}
                return True
            interface = self.get_interface()
            try:
                interface.supervisor.shutdown()
            except Exception:
                if raise_on_error:
                    raise
                return False
            lock.update_on_release = {"$set": {"running_runner": None}}
            return True

    def wait_start(self, timeout: int = 30) -> None:
        """
        Wait for all processes of the daemon to start.

        Parameters
        ----------
        timeout
            Maximum time in seconds to wait for all processes to start.
        """
        time_limit = time.time() + timeout
        while True:
            processes_info = self.get_processes_info()
            all_started = True
            for name, proc_info in processes_info.items():
                if proc_info["state"] not in RUNNING_STATES:
                    raise DaemonError(f"Process {name} is not in a running state")
                if proc_info["state"] != ProcessStates.RUNNING:
                    all_started = False
                    break
            if all_started:
                break
            if time.time() > time_limit:
                raise DaemonError(
                    f"The processes did not start within {timeout} seconds"
                )
            time.sleep(2)

    def foreground_processes(
        self,
        processes_names: list | None = None,
        print_function: Callable | None = None,
    ) -> None:
        """
        Attach to the foreground of the processes of the daemon.

        Parameters
        ----------
        processes_names
            Names of the processes to attach to. If None, all processes are
            attached to.
        print_function
            A function to use to print the output of the processes. If None,
            the default print function is used.
        """
        processes_info = self.get_processes_info()
        if processes_names is None:
            processes_names = [pn for pn in processes_info if pn != "supervisord"]

        for name in processes_names:
            if name not in processes_info:
                raise ValueError(
                    f"Process with name {name} is not among the available processes"
                )
            if processes_info[name]["state"] != ProcessStates.RUNNING:
                raise RuntimeError(
                    f"Process {name} is not running. Cannot attach to it."
                )
            self.foreground_process(name, print_function)

    def foreground_process(self, name, print_function: Callable | None = None) -> None:
        """
        Attach to the foreground of one process based on the process name.

        Parameters
        ----------
        name
            Name of the process to attach to.
        print_function
            A function to use to print the output of the process. If None,
            the default print function is used.
        """
        # This is adapted from supervisor.supervisorctl.DefaultControllerPlugin.do_fg
        a = None
        if print_function is None:
            print_function = print
        try:
            ctl = self.get_controller()

            supervisor = ctl.get_supervisor()

            print_function(f"Entering foreground for process {name} (CTRL+C to exit)")

            # this thread takes care of the output/error messages
            a = fgthread(name, ctl)
            a.start()

            # this takes care of the user input
            while True:
                # Always avoid echoing the output. It may not be a password, but since
                # the daemon process cannot control the choice here, it is safer to
                # hide everything
                # inp = raw_input() + '\n'
                inp = getpass.getpass("") + "\n"
                try:
                    supervisor.sendProcessStdin(name, inp)
                except xmlrpclib.Fault as e:
                    if e.faultCode == xmlrpc.Faults.NOT_RUNNING:
                        print_function("Process got killed")
                    else:
                        print_function("ERROR: " + str(e))
                    print_function("Exiting foreground")
                    a.kill()
                    return

                info = supervisor.getProcessInfo(name)
                if info["state"] != states.ProcessStates.RUNNING:
                    print_function("Process got killed")
                    print_function("Exiting foreground")
                    a.kill()
                    return
        except (KeyboardInterrupt, EOFError):
            print_function("Exiting foreground")
            if a:
                a.kill()

    def _get_runner_info(self) -> dict:
        """
        Generate a dictionary with the information about the runner and
        the system where it is being executed.
        """
        try:
            user = os.getlogin()
        except OSError:
            user = os.environ.get("USER", None)
        # Note that this approach may give a different MAC address for the
        # same machine, if more than one network device is present (this
        # may also include local virtual machines). Consider replacing this
        # with a more stable choice (e.g. the first one in alphabetical order,
        # excluding virtual interfaces)
        mac_address = None
        found = False
        for addrs in psutil.net_if_addrs().values():
            if found:
                break
            for addr in addrs:
                if (
                    addr.family == psutil.AF_LINK
                    and addr.address != "00:00:00:00:00:00"
                ):
                    mac_address = addr.address
                    found = True
                    break
        return {
            "processes_info": self.get_processes_info(),
            "hostname": socket.gethostname(),
            "mac_address": mac_address,
            "user": user,
            "daemon_dir": self.project.daemon_dir,
            "project_name": self.project.name,
            "start_time": datetime.datetime.now(),
            "last_pinged": datetime.datetime.now(),
            "runner_options": self.project.runner.model_dump(mode="json"),
        }

    def get_controller(self) -> Controller:
        """
        Return a Supervisor `supervisor.supervisorctl.Controller`
        connected to the supervisord configured in the project.
        """
        options = ClientOptions()
        args = ["-c", self.conf_filepath]
        options.realize(args, doc="")
        return Controller(options)

    @contextlib.contextmanager
    def lock_runner_doc(
        self, allow_missing: bool = False
    ) -> Generator[MongoLock, None, None]:
        """
        Context manager to lock a document in the auxiliary collection that is used to keep
        track of the currently running runner.

        Parameters
        ----------
        allow_missing
            Determine if it is acceptable that the document is not present in the DB.
            If not, raise an error. Mainly present to handle databases that have been
            created with version <=0.1.4.
        """
        db_filter = {"running_runner": {"$exists": True}}
        with self.job_controller.lock_auxiliary(
            filter=db_filter, get_locked_doc=True
        ) as lock:
            doc = lock.locked_document
            if not doc:
                if lock.unavailable_document:
                    raise RunnerLockedError.from_runner_doc(lock.unavailable_document)
                if not allow_missing:
                    raise ValueError(
                        "No daemon runner document.\n"
                        "The auxiliary collection does not contain information about running daemon. "
                        "Your database was likely set up or reset using an old version of Jobflow Remote. "
                        'You can upgrade the database using the command "jf admin upgrade".'
                    )
            yield lock

    def _check_running_runner(
        self, doc: dict, raise_on_error: bool = False
    ) -> str | None:
        """
        Check if the information in the DB about the running runner matches the information about the local machine.

        Parameters
        ----------
        doc
            The document from the DB with the information about the running runner.
        raise_on_error
            If True, raise an error if the information does not match, otherwise return a string with the error message.

        Returns
        -------
        str | None
            If the information matches, return None. Otherwise, return a string with the error message.
        """
        if not doc["running_runner"]:
            return None

        db_data = doc["running_runner"]

        local_data = self._get_runner_info()
        # not testing on the MAC address as it may change due to identifying
        # different network devices on the same machine or changing in
        # cloud VMs.
        data_to_check = [
            "hostname",
            "project_name",
            "user",
            "daemon_dir",
        ]
        for data in data_to_check:
            if local_data[data] != db_data[data]:
                break
        else:
            logger.info(
                "The DB reports that there is a running runner and it corresponds to this machine"
            )
            return None

        error = (
            "A daemon runner process may be running on a different machine.\n"
            "Here is the information retrieved from the database:\n"
            f"- hostname: {db_data['hostname']}\n"
            f"- project_name: {db_data['project_name']}\n"
            f"- start_time: {db_data['start_time']}\n"
            f"- last_pinged: {db_data['last_pinged']}\n"
            f"- daemon_dir: {db_data['daemon_dir']}\n"
            f"- user: {db_data['user']}"
        )
        if raise_on_error:
            raise RunningDaemonError(error)
        return error

    def check_matching_runner(
        self, raise_on_error: bool = False, allow_missing: bool = False
    ) -> str | None:
        """
        Checks if the information stored in the runner document in the auxiliary collection
        matches the current machine.

        Parameters
        ----------
        raise_on_error
            If True and the information does not match, raise an error.
        allow_missing
            If True, return None instead of raising an error if the document is not present.

        Returns
        -------
        str | None
            If the information matches, return None. Otherwise, return a string with the error message.
        """
        with self.lock_runner_doc(allow_missing=allow_missing) as lock:
            doc = lock.locked_document
            if allow_missing and not doc:
                return None
            return self._check_running_runner(doc, raise_on_error=raise_on_error)

    def parse_config_file(self) -> dict:
        """
        Parses a supervisord config file and returns the extracted input variables.
        This includes values for `num_procs_transfer`, `num_procs_complete`, `single`,
        `log_level`, `nodaemon`, and `connect_interactive`.

        Returns
        -------
        dict
            Dictionary containing the extracted input variables.
        """
        variables = {
            "num_procs_transfer": 1,
            "num_procs_complete": 1,
            "single": False,
            "log_level": "info",
            "nodaemon": False,
            "connect_interactive": False,
        }

        with open(self.conf_filepath) as f:
            content = f.read()

        # Check if this is the 'single' configuration
        if "[program:runner_daemon]" in content:
            variables["single"] = True
        else:
            variables["single"] = False

            # Extract num_procs_transfer and num_procs_complete if they exist in the split configuration
            match_transfer = re.search(r"numprocs\s*=\s*(\d+)", content, re.MULTILINE)
            if match_transfer:
                variables["num_procs_transfer"] = int(match_transfer.group(1))

            match_complete = re.search(
                r"\[program:runner_daemon_complete\][^[]*numprocs\s*=\s*(\d+)",
                content,
                re.MULTILINE,
            )
            if match_complete:
                variables["num_procs_complete"] = int(match_complete.group(1))

        # Extract loglevel
        match_loglevel = re.search(r"loglevel\s*=\s*(\w+)", content)
        if match_loglevel:
            variables["log_level"] = match_loglevel.group(1)

        # Extract nodaemon
        match_nodaemon = re.search(r"nodaemon\s*=\s*(\w+)", content)
        if match_nodaemon:
            variables["nodaemon"] = match_nodaemon.group(1) == "true"

        # Extract connect_interactive
        if "--connect-interactive" in content:
            variables["connect_interactive"] = True

        return variables



def test_version() -> None:
    from jobflow_remote import __version__

    assert __version__.startswith("0."), f"unexpected {__version__=}"


def test_imports() -> None:
    """This test triggers all the top-level imports by importing
    the global `SETTINGS`.

    """
    from jobflow_remote import SETTINGS  # noqa: F401



from __future__ import annotations

import os
import shutil
import subprocess
import warnings
from pathlib import Path

from monty.os import cd

from jobflow_remote.remote.host.base import BaseHost


class LocalHost(BaseHost):
    def __init__(self, timeout_execute: int = None, sanitize: bool = False) -> None:
        self.timeout_execute = timeout_execute
        super().__init__(sanitize=sanitize)

    def __eq__(self, other):
        return isinstance(other, LocalHost)

    def execute(
        self,
        command: str | list[str],
        workdir: str | Path | None = None,
        timeout: int | None = None,
    ):
        """Execute the given command on the host.

        Note that the command is executed with shell=True, so commands can
        be exposed to command injection. Consider whether to escape part of
        the input if it comes from external users.

        Parameters
        ----------
        command: str or list of str
            Command to execute, as a str or list of str
        workdir: str or None
            path where the command will be executed.
        timeout
            Timeout for the execution of the commands.

        Returns
        -------
        stdout : str
            Standard output of the command
        stderr : str
            Standard error of the command
        exit_code : int
            Exit code of the command.
        """
        if isinstance(command, (list, tuple)):
            command = " ".join(command)
        command = self.sanitize_command(command)
        workdir = str(workdir) if workdir else Path.cwd()
        timeout = timeout or self.timeout_execute
        with cd(workdir):
            proc = subprocess.run(
                command, capture_output=True, shell=True, timeout=timeout, check=False
            )
        stdout = self.sanitize_output(proc.stdout.decode())
        stderr = self.sanitize_output(proc.stderr.decode())
        return stdout, stderr, proc.returncode

    def mkdir(
        self, directory: str | Path, recursive: bool = True, exist_ok: bool = True
    ) -> bool:
        try:
            Path(directory).mkdir(parents=recursive, exist_ok=exist_ok)
        except OSError:
            return False
        return True

    def write_text_file(self, filepath, content) -> None:
        Path(filepath).write_text(content)

    def connect(self) -> None:
        pass

    def close(self) -> bool:
        return True

    @property
    def is_connected(self) -> bool:
        return True

    def put(self, src, dst) -> None:
        is_file_like = hasattr(src, "read") and callable(src.read)

        src_base = getattr(src, "name", None) if is_file_like else os.path.basename(src)

        if Path(dst).is_dir():
            if src_base:
                dst = Path(dst, src_base)
            elif is_file_like:
                raise ValueError(
                    "could not determine the file name and dst is a folder"
                )

        if is_file_like:
            with open(dst, "wb") as f:
                f.write(src.read())
        else:
            self.copy(src, dst)

    def get(self, src, dst) -> None:
        is_file_like = hasattr(dst, "write") and callable(dst.write)

        if is_file_like:
            with open(src, "rb") as f:
                dst.write(f.read())
        else:
            self.copy(src, dst)

    def copy(self, src, dst) -> None:
        shutil.copy(src, dst)

    def listdir(self, path: str | Path) -> list[str]:
        try:
            return os.listdir(path)
        except FileNotFoundError:
            return []

    def remove(self, path: str | Path) -> None:
        os.remove(path)

    def rmtree(self, path: str | Path, raise_on_error: bool = False) -> bool:
        """Recursively delete a directory tree on a local host.

        It is intended to remove an entire directory tree, including all files
        and subdirectories, on this local host.

        Parameters
        ----------
        path : str or Path
            The path to the directory tree to be removed.

        raise_on_error : bool
            If set to `False` (default), errors will be ignored, and the method will
            attempt to continue removing remaining files and directories.
            Otherwise, any errors encountered during the removal process
            will raise an exception.

        Returns
        -------
        bool
            True if the directory tree was successfully removed, False otherwise.
        """
        removed = True

        def onerror(func, dir_path, err):
            nonlocal removed
            removed = False
            warnings.warn(f"Error while deleting folder {path}: {err[1]}", stacklevel=2)

        shutil.rmtree(path, onerror=onerror if not raise_on_error else None)

        return removed



from __future__ import annotations

from typing import TYPE_CHECKING

from jobflow_remote.jobs.state import JobState

if TYPE_CHECKING:
    from networkx import DiGraph

    from jobflow_remote.jobs.data import FlowInfo


def get_graph(flow: FlowInfo, label: str = "name") -> DiGraph:
    import networkx as nx

    graph = nx.DiGraph()

    ids_mapping = flow.ids_mapping

    # Add nodes
    for job_prop in flow.iter_job_prop():
        db_id = job_prop["db_id"]
        job_prop["label"] = job_prop[label]
        # change this as the "name" is used in jobflow's graph plotting util
        job_prop["job_name"] = job_prop.pop("name")
        graph.add_node(db_id, **job_prop)

    # Add edges based on parents
    for child_node, parents in zip(flow.db_ids, flow.parents):
        for parent_uuid in parents:
            for parent_node in ids_mapping[parent_uuid].values():
                graph.add_edge(parent_node, child_node)

    return graph


def get_graph_elements(flow: FlowInfo):
    ids_mapping = flow.ids_mapping

    nodes = {}
    hosts = {}
    for job_prop in flow.iter_job_prop():
        db_id = job_prop["db_id"]
        nodes[db_id] = job_prop
        hosts[job_prop["db_id"]] = job_prop["hosts"]

    # edges based on parents
    edges = [
        (parent_node, child_node)
        for child_node, parents in zip(flow.db_ids, flow.parents)
        for parent_uuid in parents
        for parent_node in ids_mapping[parent_uuid].values()
    ]

    # connections between replacements
    replace_edges = []
    for dct in flow.ids_mapping.values():
        # skip index 1
        for job_index in sorted(dct)[1:]:
            job_db_id = dct[job_index]
            previous_index_db_id = dct[job_index - 1]
            # likely never entering here, because if a single job is used as a
            # "replace" it still goes through a get_flow in jobflow in prepare_replace.
            if hosts[job_db_id][0] == hosts[previous_index_db_id][0]:
                replace_edges.append((previous_index_db_id, job_db_id))
            else:
                replace_edges.append((previous_index_db_id, hosts[job_db_id][0]))

    return nodes, edges, hosts, replace_edges


def plot_dash(flow: FlowInfo):
    nodes, edges, hosts, replace_edges = get_graph_elements(flow)

    import dash_cytoscape as cyto
    from dash import Dash, Input, Output, callback, html

    app = Dash(f"{flow.name} - {flow.flow_id}")

    elements: list[dict] = []

    # parent elements
    hosts_hierarchy = {}
    jobs_inner_hosts = {}
    hosts_set = set()
    for db_id, job_hosts in hosts.items():
        job_hosts = list(reversed(job_hosts))  # noqa: PLW2901
        if len(job_hosts) < 2:
            continue
        for i, host in enumerate(job_hosts[1:-1], 1):
            hosts_hierarchy[job_hosts[i + 1]] = host

        hosts_set.update(job_hosts[1:])
        jobs_inner_hosts[db_id] = job_hosts[-1]

    elements += [
        {"data": {"id": host, "parent": hosts_hierarchy.get(host)}}
        for host in hosts_set
    ]

    for db_id, node_info in nodes.items():
        node_info["id"] = str(db_id)
        node_info["label"] = node_info["name"]
        node_info["parent"] = jobs_inner_hosts.get(db_id)
        elements.append(
            {
                "data": node_info,
            }
        )

    elements += [
        {"data": {"source": str(edge[0]), "target": str(edge[1])}} for edge in edges
    ]

    for edge in replace_edges:
        elements.append(
            {
                "data": {"source": str(edge[0]), "target": str(edge[1])},
                "classes": "dashed",
            }
        )

    stylesheet: list[dict] = [
        {
            "selector": f'[state = "{state}"]',
            "style": {
                "background-color": color,
            },
        }
        for state, color in COLOR_MAPPING.items()
    ]
    stylesheet.append(
        {
            "selector": "node",
            "style": {
                "label": "data(name)",
            },
        }
    )
    stylesheet.append(
        {
            "selector": "node:parent",
            "style": {
                "background-opacity": 0.2,
                "background-color": "#2B65EC",
                "border-color": "#2B65EC",
            },
        }
    )

    stylesheet.append(
        {
            "selector": ".dashed",
            "style": {
                "line-style": "dashed",
            },
        }
    )

    app.layout = html.Div(
        [
            cyto.Cytoscape(
                id="flow-graph",
                layout={"name": "breadthfirst", "directed": True},
                # layout={'name': 'cose'},
                style={"width": "100%", "height": "500px"},
                elements=elements,
                stylesheet=stylesheet,
            ),
            html.P(id="job-info-output"),
        ]
    )

    @callback(
        Output("job-info-output", "children"), Input("flow-graph", "mouseoverNodeData")
    )
    def displayTapNodeData(data):
        if data:
            return str(data)
        return None

    app.run(debug=True)


def get_mermaid(flow: FlowInfo, show_subflows: bool = True):
    nodes, edges, hosts, replace_edges = get_graph_elements(flow)
    from monty.collections import tree

    hosts_hierarchy = tree()
    for db_id, job_hosts in hosts.items():
        d = hosts_hierarchy
        for host in reversed(job_hosts):
            d = d[host]
        d[db_id] = None

    lines = ["flowchart TD"]

    # add style classes
    for state, color in COLOR_MAPPING.items():
        # this could be optimised by compressing in one line and using a
        # same class for states with the same color
        lines.append(f"    classDef {state} fill:{color}")

    # add edges
    for parent_db_id, child_db_id in edges:
        parent = nodes[parent_db_id]
        child = nodes[child_db_id]
        line = (
            f"    {parent_db_id}({parent['name']}) --> {child_db_id}({child['name']})"
        )
        lines.append(line)

    # add replace edges
    for parent_db_id, child_id in replace_edges:
        line = f"    {parent_db_id}({nodes[parent_db_id]['name']}) -.-> {child_id}"
        lines.append(line)

    subgraph_styles = []

    # add subgraphs
    def add_subgraph(nested_hosts_hierarchy, indent_level=0) -> None:
        prefix = "    " * indent_level if show_subflows else "    "

        for ref_id in sorted(nested_hosts_hierarchy, key=lambda x: str(x)):
            subhosts = nested_hosts_hierarchy[ref_id]
            if subhosts:
                if indent_level > 0 and show_subflows:
                    # don't  put any title
                    lines.append(f"{prefix}subgraph {ref_id}[ ]")
                    subgraph_styles.append(
                        f"    style {ref_id} fill:#2B65EC,opacity:0.2"
                    )

                add_subgraph(subhosts, indent_level=indent_level + 1)

                if indent_level > 0 and show_subflows:
                    lines.append(f"{prefix}end")
            else:
                job = nodes[ref_id]
                lines.append(f"{prefix}{ref_id}:::{job['state'].value}")

    add_subgraph(hosts_hierarchy)
    lines.extend(subgraph_styles)

    return "\n".join(lines)


BLUE_COLOR = "#5E6BFF"
RED_COLOR = "#fC3737"
COLOR_MAPPING = {
    JobState.WAITING.value: "#aaaaaa",
    JobState.READY.value: "#DAF7A6",
    JobState.CHECKED_OUT.value: BLUE_COLOR,
    JobState.UPLOADED.value: BLUE_COLOR,
    JobState.SUBMITTED.value: BLUE_COLOR,
    JobState.RUNNING.value: BLUE_COLOR,
    JobState.TERMINATED.value: BLUE_COLOR,
    JobState.DOWNLOADED.value: BLUE_COLOR,
    JobState.REMOTE_ERROR.value: RED_COLOR,
    JobState.COMPLETED.value: "#47bf00",
    JobState.FAILED.value: RED_COLOR,
    JobState.PAUSED.value: "#EAE200",
    JobState.STOPPED.value: RED_COLOR,
    JobState.USER_STOPPED.value: RED_COLOR,
    JobState.BATCH_SUBMITTED.value: BLUE_COLOR,
    JobState.BATCH_RUNNING.value: BLUE_COLOR,
}



from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime
from typing import TYPE_CHECKING

from jobflow_remote.jobs.data import JobInfo, projection_job_info
from jobflow_remote.jobs.state import FlowState, JobState

if TYPE_CHECKING:
    from jobflow_remote import JobController


@dataclass
class JobTrends:
    """
    Trends of job states over time.
    """

    interval: str
    dates: list[str]
    completed: list[int]
    failed: list[int]
    remote_error: list[int]
    timezone: str

    @property
    def num_intervals(self) -> int:
        return len(self.dates)


@dataclass
class JobsReport:
    """
    A report of the job states.
    """

    state_counts: dict[JobState, int] = field(default_factory=dict)
    trends: JobTrends | None = None
    longest_running: list[JobInfo] = field(default_factory=list)
    worker_utilization: dict[str, int] = field(default_factory=dict)

    @property
    def running(self) -> int:
        """Returns the count of running jobs."""
        return self.state_counts.get(JobState.RUNNING, 0)

    @property
    def completed(self) -> int:
        """Returns the count of completed jobs."""
        return self.state_counts.get(JobState.COMPLETED, 0)

    @property
    def error(self) -> int:
        """Returns the sum of failed, remote error, and paused jobs (i.e., error states)."""
        return sum(
            self.state_counts.get(state, 0)
            for state in [JobState.FAILED, JobState.REMOTE_ERROR]
        )

    @property
    def active(self) -> int:
        """Returns the sum of failed, remote error, and paused jobs (i.e., error states)."""
        return sum(
            self.state_counts.get(state, 0)
            for state in [
                JobState.CHECKED_OUT,
                JobState.UPLOADED,
                JobState.UPLOADED,
                JobState.SUBMITTED,
                JobState.RUNNING,
                JobState.TERMINATED,
                JobState.DOWNLOADED,
                JobState.BATCH_SUBMITTED,
                JobState.BATCH_RUNNING,
            ]
        )

    @classmethod
    def generate_report(
        cls,
        job_controller: JobController,
        interval: str = "days",
        num_intervals: int | None = None,
        timezone: str = "UTC",
    ) -> JobsReport:
        """
        Generates a report of the job states.

        Parameters
        ----------
        job_controller
            The JobController instance to generate the report from.
        interval
            The interval of the trends for the report.
        num_intervals
            The number of intervals to consider.
        timezone
            The timezone to use for the report.

        Returns
        -------
        JobsReport
            A report of the job states.
        """
        now = datetime.utcnow()

        state_counts = job_controller.count_jobs_states(list(JobState))

        # Job trends over time
        states_trends = [JobState.COMPLETED, JobState.FAILED, JobState.REMOTE_ERROR]
        trends_dict = job_controller.get_trends(
            states=states_trends,
            interval=interval,
            num_intervals=num_intervals,
            interval_timezone=timezone,
        )
        trends_dates = sorted(trends_dict)
        trends = JobTrends(
            interval=interval,
            dates=trends_dates,
            completed=[trends_dict[d][JobState.COMPLETED] for d in trends_dates],
            failed=[trends_dict[d][JobState.FAILED] for d in trends_dates],
            remote_error=[trends_dict[d][JobState.REMOTE_ERROR] for d in trends_dates],
            timezone=timezone,
        )

        # Longest running jobs (Top 5)
        projection_longest: dict = {k: 1 for k in projection_job_info}
        projection_longest["duration"] = {"$subtract": [now, "$start_time"]}
        pipeline_longest = [
            {"$match": {"state": "RUNNING"}},
            {"$project": projection_longest},
            {"$sort": {"duration": -1}},
            {"$limit": 5},
        ]
        longest_running_result = job_controller.jobs.aggregate(pipeline_longest)
        longest_running = [
            JobInfo.from_query_output(doc) for doc in longest_running_result
        ]

        # Worker utilization (number of jobs assigned to each worker)
        pipeline_worker_utilization = [
            {"$group": {"_id": "$worker", "count": {"$sum": 1}}},
            {"$sort": {"count": -1}},
        ]
        worker_utilization_result = job_controller.jobs.aggregate(
            pipeline_worker_utilization
        )
        worker_utilization = {
            doc["_id"]: doc["count"] for doc in worker_utilization_result
        }

        return cls(
            state_counts=state_counts,
            trends=trends,
            longest_running=longest_running,
            worker_utilization=worker_utilization,
        )


@dataclass
class FlowTrends:
    """
    Trends of flow states over time.
    """

    interval: str
    dates: list[str]
    completed: list[int]
    failed: list[int]
    timezone: str

    @property
    def num_intervals(self) -> int:
        return len(self.dates)


@dataclass
class FlowsReport:
    """
    A report of the flow states.
    """

    state_counts: dict[FlowState, int]
    trends: FlowTrends

    @property
    def running(self) -> int:
        """Returns the count of running flows."""
        return self.state_counts.get(FlowState.RUNNING, 0)

    @property
    def completed(self) -> int:
        """Returns the count of completed flows."""
        return self.state_counts.get(FlowState.COMPLETED, 0)

    @property
    def error(self) -> int:
        """Returns the count of failed flows."""
        return self.state_counts.get(FlowState.FAILED, 0)

    @classmethod
    def generate_report(
        cls,
        job_controller: JobController,
        interval: str = "days",
        num_intervals: int = None,
        timezone: str = "UTC",
    ):
        """
        Generates a report of the flow states.

        Parameters
        ----------
        job_controller
            The JobController instance to generate the report from.
        interval
            The interval of the trends for the report.
        num_intervals
            The number of intervals to consider.
        timezone
            The timezone to use for the report.

        Returns
        -------
        FlowsReport
            A report of the flow states.
        """
        state_counts = job_controller.count_flows_states(list(FlowState))

        trends_dict = job_controller.get_trends(
            states=[FlowState.COMPLETED, FlowState.FAILED],
            interval=interval,
            num_intervals=num_intervals,
            interval_timezone=timezone,
        )
        trends_dates = sorted(trends_dict)
        trends = FlowTrends(
            interval=interval,
            dates=trends_dates,
            completed=[trends_dict[d][FlowState.COMPLETED] for d in trends_dates],
            failed=[trends_dict[d][FlowState.FAILED] for d in trends_dates],
            timezone=timezone,
        )

        # Create report instance
        return cls(state_counts=state_counts, trends=trends)



import abc
import logging
import traceback
from enum import Enum
from pathlib import Path
from typing import Annotated, Any, Literal, Optional, Union

from jobflow import JobStore
from maggma.stores import MongoStore
from pydantic import BaseModel, ConfigDict, Field, ValidationInfo, field_validator
from qtoolkit.io import BaseSchedulerIO, scheduler_mapping

from jobflow_remote.remote.host import BaseHost, LocalHost, RemoteHost
from jobflow_remote.utils.data import store_from_dict

DEFAULT_JOBSTORE = {"docs_store": {"type": "MemoryStore"}}


class RunnerOptions(BaseModel):
    """Options to tune the execution of the Runner."""

    delay_checkout: int = Field(
        30,
        description="Delay between subsequent execution of the checkout from database (seconds)",
    )
    delay_check_run_status: int = Field(
        30,
        description="Delay between subsequent execution of the checking the status of "
        "jobs that are submitted to the scheduler (seconds)",
    )
    delay_advance_status: int = Field(
        30,
        description="Delay between subsequent advancement of the job's remote state (seconds)",
    )
    delay_refresh_limited: int = Field(
        600,
        description="Delay between subsequent refresh from the DB of the number of submitted "
        "and running jobs (seconds). Only used if a worker with max_jobs is present",
    )
    delay_update_batch: int = Field(
        60,
        description="Delay between subsequent refresh from the DB of the number of submitted "
        "and running jobs (seconds). Only used if a batch worker is present",
    )
    delay_ping_db: int = Field(
        7200,
        description="Delay between subsequent pings to the running runner document.",
    )
    lock_timeout: Optional[int] = Field(
        86400,
        description="Time to consider the lock on a document expired and can be overridden (seconds)",
    )
    delete_tmp_folder: bool = Field(
        default=True,
        description="Whether to delete the local temporary folder after a job has completed",
    )
    max_step_attempts: int = Field(
        3,
        description="Maximum number of attempt performed before failing an "
        "advancement of a remote state",
    )
    delta_retry: tuple[int, ...] = Field(
        (30, 300, 1200),
        description="List of increasing delay between subsequent attempts when the "
        "advancement of a remote step fails",
    )

    def get_delta_retry(self, step_attempts: int) -> int:
        """
        The time to wait before retrying a failed advancement of the remote state,
        based on the number of attempts.

        If exceeding the size of the list delta_retry, the last value is returned.

        Parameters
        ----------
        step_attempts
            The number of attempts advancing a remote state.

        Returns
        -------
            The delay in seconds.
        """
        ind = min(step_attempts, len(self.delta_retry)) - 1
        return self.delta_retry[ind]

    model_config = ConfigDict(extra="forbid")


class LogLevel(str, Enum):
    """Enumeration of logging level."""

    ERROR = "error"
    WARN = "warn"
    INFO = "info"
    DEBUG = "debug"

    def to_logging(self) -> int:
        """
        Helper converter to python logging values.

        Returns
        -------
        The int corresponding to python logging value
        """
        return {
            LogLevel.ERROR: logging.ERROR,
            LogLevel.WARN: logging.WARNING,
            LogLevel.INFO: logging.INFO,
            LogLevel.DEBUG: logging.DEBUG,
        }[self]


class BatchConfig(BaseModel):
    """
    Configuration for execution of batch jobs.

    Allows to execute multiple Jobs in a single process executed on the
    worker (e.g. SLURM job).
    """

    jobs_handle_dir: Path = Field(
        description="Absolute path to a folder that will be used to store information to share with the jobs being executed"
    )
    work_dir: Path = Field(
        description="Absolute path to a folder where the batch jobs will be executed. This refers to the jobs submitted"
        "to the queue. Jobflow's Job will still be executed in the standard folders."
    )
    max_jobs_per_batch: Optional[int] = Field(
        None, description="Maximum number of jobs executed in a single batch process"
    )
    max_wait: Optional[int] = Field(
        60,
        description="Maximum time to wait before stopping if no new jobs are available to run (seconds)",
    )
    max_time: Optional[int] = Field(
        None,
        description="Maximum time after which a job will not start more jobs (seconds). To help avoid hitting the walltime",
    )
    parallel_jobs: Optional[int] = Field(
        None, description="Number of jobs executed in parallel in the same process"
    )
    model_config = ConfigDict(extra="forbid")


class WorkerBase(BaseModel):
    """Base class defining the common field for the different types of Worker."""

    type: str = Field(
        description="The discriminator field to determine the worker type"
    )

    scheduler_type: str = Field(
        description="Type of the scheduler. Depending on the values supported by QToolKit"
    )
    work_dir: Path = Field(
        description="Absolute path of the directory of the worker where subfolders for "
        "executing the calculation will be created"
    )
    resources: Optional[dict] = Field(
        None,
        description="A dictionary defining the default resources requested to the "
        "scheduler. Used to fill in the QToolKit template",
    )
    pre_run: Optional[str] = Field(
        None,
        description="String with commands that will be executed before the execution of the Job",
    )
    post_run: Optional[str] = Field(
        None,
        description="String with commands that will be executed after the execution of the Job",
    )
    execution_cmd: Optional[str] = Field(
        None,
        description="String with commands to execute the Job on the worker. By default will be "
        "set to `jf -fe execution run {}`. The `{}` part will be used to insert the path to "
        "the execution directory and it is mandatory. Change only for specific needs (e.g. the"
        "jf command needs to be executed in a container).",
    )
    timeout_execute: int = Field(
        60,
        description="Timeout for the execution of the commands in the worker "
        "(e.g. submitting a job)",
    )
    max_jobs: Optional[int] = Field(
        None,
        description="The maximum number of jobs that can be submitted to the queue.",
        ge=0,
    )
    batch: Optional[BatchConfig] = Field(
        None,
        description="Options for batch execution. If define the worker will be considered a batch worker",
    )
    scheduler_username: Optional[str] = Field(
        None,
        description="If defined, the list of jobs running on the worker will be fetched based on the"
        "username instead that from the list of job ids. May be necessary for some "
        "scheduler_type (e.g. SGE)",
    )
    sanitize_command: bool = Field(
        default=False,
        description="Sanitize the output of commands in case of failures due to spurious text produced"
        "by the worker shell.",
    )
    delay_download: Optional[int] = Field(
        default=None,
        description="Amount of seconds to wait to start the download after the Runner marked a Job "
        "as TERMINATED. To account for delays in the writing of the file on the worker file system"
        " (e.g. NFS).",
    )
    model_config = ConfigDict(extra="forbid")

    @field_validator("scheduler_type")
    @classmethod
    def check_scheduler_type(cls, scheduler_type: str) -> str:
        """Validator to set the default of scheduler_type."""
        if scheduler_type not in scheduler_mapping:
            raise ValueError(f"Unknown scheduler type {scheduler_type}")
        return scheduler_type

    @field_validator("work_dir")
    @classmethod
    def check_work_dir(cls, v) -> Path:
        if not v.is_absolute():
            raise ValueError("`work_dir` must be an absolute path")
        return v

    @field_validator("execution_cmd")
    @classmethod
    def check_execution_cmd(cls, v) -> Optional[str]:
        if v is not None and "{}" not in v:
            raise ValueError(
                "`execution_cmd` must contain a '{}' part to allow setting "
                "the execution folder at runtime"
            )
        return v

    def get_scheduler_io(self) -> BaseSchedulerIO:
        """
        Get the BaseSchedulerIO from QToolKit depending on scheduler_type.

        Returns
        -------
        The instance of the scheduler_type.
        """
        if self.scheduler_type not in scheduler_mapping:
            raise ConfigError(f"Unknown scheduler type {self.scheduler_type}")
        return scheduler_mapping[self.scheduler_type]()

    @abc.abstractmethod
    def get_host(self) -> BaseHost:
        """Return the Host object used in the Worker."""

    @property
    @abc.abstractmethod
    def cli_info(self) -> dict:
        """
        Short information about the worker to be displayed in the command line
        interface.

        Returns
        -------
        A dictionary with the Worker short information.
        """

    @property
    def is_local(self) -> bool:
        return self.type == "local"


class LocalWorker(WorkerBase):
    """
    Worker representing the local host.

    Executes command directly.
    """

    type: Literal["local"] = Field(
        "local", description="The discriminator field to determine the worker type"
    )

    def get_host(self) -> BaseHost:
        """
        Return the LocalHost.

        Returns
        -------
        The LocalHost.
        """
        return LocalHost(
            timeout_execute=self.timeout_execute, sanitize=self.sanitize_command
        )

    @property
    def cli_info(self) -> dict:
        """
        Short information about the worker to be displayed in the command line
        interface.

        Returns
        -------
        A dictionary with the Worker short information.
        """
        return dict(
            scheduler_type=self.scheduler_type,
            work_dir=self.work_dir,
        )


class ConnectionData(BaseModel):
    """
    The representation of a fabric connection.
    Mainly used in case of nested gateways.
    """

    host: str = Field(description="The host to which to connect")
    user: Optional[str] = Field(None, description="Login username")
    port: Optional[int] = Field(None, description="Port number")
    password: Optional[str] = Field(None, description="Login password")
    key_filename: Optional[Union[str, list[str]]] = Field(
        None,
        description="The filename, or list of filenames, of optional private key(s) "
        "and/or certs to try for authentication",
    )
    passphrase: Optional[str] = Field(
        None, description="Passphrase used for decrypting private keys"
    )
    gateway: Optional[Union[str, "ConnectionData"]] = Field(
        None, description="A shell command string to use as a proxy or gateway"
    )
    connect_kwargs: Optional[dict] = Field(
        None,
        description="Other keyword arguments passed to paramiko.client.SSHClient.connect",
    )

    def get_connect_kwargs(self) -> dict:
        """
        Return the fully filled connect_kwargs for Fabric.

        Returns
        -------
        The RemoteHost.
        """
        connect_kwargs = dict(self.connect_kwargs) if self.connect_kwargs else {}
        if self.password:
            connect_kwargs["password"] = self.password
        if self.key_filename:
            connect_kwargs["key_filename"] = self.key_filename
        if self.passphrase:
            connect_kwargs["passphrase"] = self.passphrase

        return connect_kwargs


class RemoteWorker(WorkerBase):
    """
    Worker representing a remote host reached through an SSH connection.

    Uses a Fabric Connection. Check Fabric documentation for more details on the
    options defining a Connection.
    """

    type: Literal["remote"] = Field(
        "remote", description="The discriminator field to determine the worker type"
    )
    host: str = Field(description="The host to which to connect")
    user: Optional[str] = Field(None, description="Login username")
    port: Optional[int] = Field(None, description="Port number")
    password: Optional[str] = Field(None, description="Login password")
    key_filename: Optional[Union[str, list[str]]] = Field(
        None,
        description="The filename, or list of filenames, of optional private key(s) "
        "and/or certs to try for authentication",
    )
    passphrase: Optional[str] = Field(
        None, description="Passphrase used for decrypting private keys"
    )
    gateway: Optional[Union[str, ConnectionData]] = Field(
        None, description="A shell command string to use as a proxy or gateway"
    )
    forward_agent: Optional[bool] = Field(
        None, description="Whether to enable SSH agent forwarding"
    )
    connect_timeout: Optional[int] = Field(
        None, description="Connection timeout, in seconds"
    )
    connect_kwargs: Optional[dict] = Field(
        None,
        description="Other keyword arguments passed to paramiko.client.SSHClient.connect",
    )
    inline_ssh_env: Optional[bool] = Field(
        None,
        description="Whether to send environment variables 'inline' as prefixes in "
        "front of command strings",
    )
    keepalive: Optional[int] = Field(
        60, description="Keepalive value in seconds passed to paramiko's transport"
    )
    shell_cmd: Optional[str] = Field(
        "bash",
        description="The shell command used to execute the command remotely. If None "
        "the command is executed directly",
    )
    login_shell: bool = Field(
        default=True,
        description="Whether to use a login shell when executing the command",
    )
    interactive_login: bool = Field(
        default=False,
        description="Whether the authentication to the host should be interactive",
    )

    def get_host(self) -> BaseHost:
        """
        Return the RemoteHost.

        Returns
        -------
        The RemoteHost.
        """
        connect_kwargs = dict(self.connect_kwargs) if self.connect_kwargs else {}
        if self.password:
            connect_kwargs["password"] = self.password
        if self.key_filename:
            connect_kwargs["key_filename"] = self.key_filename
        if self.passphrase:
            connect_kwargs["passphrase"] = self.passphrase
        return RemoteHost(
            host=self.host,
            user=self.user,
            port=self.port,
            gateway=self.gateway,
            forward_agent=self.forward_agent,
            connect_timeout=self.connect_timeout,
            connect_kwargs=connect_kwargs,
            inline_ssh_env=self.inline_ssh_env,
            timeout_execute=self.timeout_execute,
            keepalive=self.keepalive,
            shell_cmd=self.shell_cmd,
            login_shell=self.login_shell,
            interactive_login=self.interactive_login,
            sanitize=self.sanitize_command,
        )

    @property
    def cli_info(self) -> dict:
        """
        Short information about the worker to be displayed in the command line
        interface.

        Returns
        -------
        A dictionary with the Worker short information.
        """
        return dict(
            host=self.host,
            scheduler_type=self.scheduler_type,
            work_dir=self.work_dir,
        )


WorkerConfig = Annotated[Union[LocalWorker, RemoteWorker], Field(discriminator="type")]


class ExecutionConfig(BaseModel):
    """Configuration to be set before and after the execution of a Job."""

    modules: Optional[list[str]] = Field(
        None, description="list of modules to be loaded"
    )
    export: Optional[dict[str, Any]] = Field(
        None, description="dictionary with variable to be exported"
    )
    pre_run: Optional[str] = Field(
        None, description="Other commands to be executed before the execution of a job"
    )
    post_run: Optional[str] = Field(
        None, description="Commands to be executed after the execution of a job"
    )
    model_config = ConfigDict(extra="forbid")


class QueueConfig(BaseModel):
    store: dict = Field(
        default_factory=dict,
        description="Dictionary describing a maggma Store used for the queue data. "
        "Can contain the monty serialized dictionary or a dictionary with a 'type' "
        "specifying the Store subclass. Should be subclass of a MongoStore, as it "
        "requires to perform MongoDB actions. The collection is used to store the "
        "jobs",
        validate_default=True,
    )
    flows_collection: str = Field(
        "flows",
        description="The name of the collection containing information about the flows. "
        "Taken from the same database as the one defined in the store",
    )
    auxiliary_collection: str = Field(
        "jf_auxiliary",
        description="The name of the collection containing auxiliary information. "
        "Taken from the same database as the one defined in the store",
    )
    db_id_prefix: Optional[str] = Field(
        None,
        description="a string defining the prefix added to the integer ID associated "
        "to each Job in the database",
    )

    @field_validator("store")
    @classmethod
    def check_store(cls, store: dict) -> dict:
        """Check that the queue configuration could be converted to a Store."""
        if store:
            try:
                deserialized_store = store_from_dict(store)
            except Exception as e:
                raise ValueError(
                    f"error while converting queue to a maggma store. Error: {traceback.format_exc()}"
                ) from e
            if not isinstance(deserialized_store, MongoStore):
                raise ValueError(
                    "The queue store should be a subclass of a "
                    f"MongoStore: {type(deserialized_store)} instead"
                )
        return store

    model_config = ConfigDict(extra="forbid")


class Project(BaseModel):
    """The configurations of a Project."""

    name: str = Field(description="The name of the project")
    base_dir: Optional[str] = Field(
        None,
        description="The base directory containing the project related files. Default "
        "is a folder with the project name inside the projects folder",
        validate_default=True,
    )
    tmp_dir: Optional[str] = Field(
        None,
        description="Folder where remote files are copied. Default a 'tmp' folder in base_dir",
        validate_default=True,
    )
    log_dir: Optional[str] = Field(
        None,
        description="Folder containing all the logs. Default a 'log' folder in base_dir",
        validate_default=True,
    )
    daemon_dir: Optional[str] = Field(
        None,
        description="Folder containing daemon related files. Default to a 'daemon' "
        "folder in base_dir",
        validate_default=True,
    )
    log_level: LogLevel = Field(LogLevel.INFO, description="The level set for logging")
    runner: RunnerOptions = Field(
        default_factory=RunnerOptions, description="The options for the Runner"
    )
    workers: dict[str, WorkerConfig] = Field(
        default_factory=dict,
        description="A dictionary with the worker name as keys and the worker "
        "configuration as values",
    )
    queue: QueueConfig = Field(
        description="The configuration of the Store used to store the states of"
        " the Jobs and the Flows",
    )
    exec_config: dict[str, ExecutionConfig] = Field(
        default_factory=dict,
        description="A dictionary with the ExecutionConfig name as keys and the "
        "ExecutionConfig configuration as values",
    )
    jobstore: dict = Field(
        default_factory=lambda: dict(DEFAULT_JOBSTORE),
        description="The JobStore used for the output. Can contain the monty "
        "serialized dictionary or the Store in the Jobflow format",
        validate_default=True,
    )
    remote_jobstore: Optional[dict] = Field(
        None,
        description="The JobStore used for the data transfer between the Runner"
        "and the workers. Can be a string with the standard values",
    )
    metadata: Optional[dict] = Field(
        None, description="A dictionary with metadata associated to the project"
    )

    def get_jobstore(self) -> Optional[JobStore]:
        """
        Generate an instance of the JobStore based on the configuration.

        Returns
        -------
        A JobStore
        """
        if not self.jobstore:
            return None
        if self.jobstore.get("@class") == "JobStore":
            return JobStore.from_dict(self.jobstore)
        return JobStore.from_dict_spec(self.jobstore)

    def get_queue_store(self):
        """
        Generate an instance of a maggma Store based on the queue configuration.

        Returns
        -------
        A maggma Store
        """
        return store_from_dict(self.queue.store)

    def get_job_controller(self):
        from jobflow_remote.jobs.jobcontroller import JobController

        return JobController.from_project(self)

    @field_validator("base_dir")
    @classmethod
    def check_base_dir(cls, base_dir: str, info: ValidationInfo) -> str:
        """Validator to set the default of base_dir based on the project name."""
        if not base_dir:
            from jobflow_remote import SETTINGS

            return str(Path(SETTINGS.projects_folder, info.data["name"]))
        return base_dir

    @field_validator("tmp_dir")
    @classmethod
    def check_tmp_dir(cls, tmp_dir: str, info: ValidationInfo) -> str:
        """Validator to set the default of tmp_dir based on the base_dir."""
        if not tmp_dir:
            return str(Path(info.data["base_dir"], "tmp"))
        return tmp_dir

    @field_validator("log_dir")
    @classmethod
    def check_log_dir(cls, log_dir: str, info: ValidationInfo) -> str:
        """Validator to set the default of log_dir based on the base_dir."""
        if not log_dir:
            return str(Path(info.data["base_dir"], "log"))
        return log_dir

    @field_validator("daemon_dir")
    @classmethod
    def check_daemon_dir(cls, daemon_dir: str, info: ValidationInfo) -> str:
        """Validator to set the default of daemon_dir based on the base_dir."""
        if not daemon_dir:
            return str(Path(info.data["base_dir"], "daemon"))
        return daemon_dir

    @field_validator("jobstore")
    @classmethod
    def check_jobstore(cls, jobstore: dict) -> dict:
        """Check that the jobstore configuration could be converted to a JobStore."""
        if jobstore:
            try:
                if jobstore.get("@class") == "JobStore":
                    JobStore.from_dict(jobstore)
                else:
                    JobStore.from_dict_spec(jobstore)
            except Exception as e:
                raise ValueError(
                    f"error while converting jobstore to JobStore. Error: {traceback.format_exc()}"
                ) from e
        return jobstore

    @property
    def has_interactive_workers(self) -> bool:
        """
        True if any of the workers have interactive_login set to True.
        """
        for worker in self.workers.values():
            if isinstance(worker, RemoteWorker) and worker.interactive_login:
                return True
        return False

    model_config = ConfigDict(extra="forbid")


class ConfigError(Exception):
    """A generic Exception related to the configuration."""


class ProjectUndefinedError(ConfigError):
    """Exception raised if the Project has not been defined or could not be determined."""



from __future__ import annotations

import getpass
import io
import logging
import shlex
import traceback
import warnings
from contextlib import ExitStack
from pathlib import Path
from typing import Callable

import fabric
from fabric import Config
from fabric.auth import OpenSSHAuthStrategy
from paramiko.auth_strategy import AuthSource
from paramiko.ssh_exception import SSHException

from jobflow_remote.remote.host.base import BaseHost

logger = logging.getLogger(__name__)


class RemoteHost(BaseHost):
    """
    Execute commands on a remote host.
    For some commands assumes the remote can run unix.
    """

    def __init__(
        self,
        host,
        user=None,
        port=None,
        config=None,
        gateway=None,
        forward_agent=None,
        connect_timeout=None,
        connect_kwargs=None,
        inline_ssh_env=None,
        timeout_execute=None,
        keepalive=60,
        shell_cmd="bash",
        login_shell=True,
        retry_on_closed_connection=True,
        interactive_login=False,
        sanitize: bool = False,
    ) -> None:
        self.host = host
        self.user = user
        self.port = port
        self.config = config
        self.gateway = gateway
        self.forward_agent = forward_agent
        self.connect_timeout = connect_timeout
        self.connect_kwargs = connect_kwargs
        self.inline_ssh_env = inline_ssh_env
        self.timeout_execute = timeout_execute
        self.keepalive = keepalive
        self.shell_cmd = shell_cmd
        self.login_shell = login_shell
        self.retry_on_closed_connection = retry_on_closed_connection
        self._interactive_login = interactive_login
        self._create_connection()
        super().__init__(sanitize=sanitize)

    def _create_connection(self) -> None:
        if self.interactive_login:
            # if auth_timeout is not explicitly set, use a larger value than
            # the default to avoid the timeout while user get access to the process
            connect_kwargs = dict(self.connect_kwargs) if self.connect_kwargs else {}
            if "auth_timeout" not in connect_kwargs:
                connect_kwargs["auth_timeout"] = 120
            config = self.config
            self._connection = self._get_single_connection(
                host=self.host,
                user=self.user,
                port=self.port,
                config=config,
                gateway=self.gateway,
                connect_kwargs=connect_kwargs,
            )

            # if the authentication is ssh-key + OTP paramiko already
            # handles it. Don't use the alternative strategy.
            if not self._connection.connect_kwargs.get("key_filename"):
                if not config:
                    config = Config()
                    config.authentication.strategy_class = InteractiveAuthStrategy

                self._connection = self._get_single_connection(
                    host=self.host,
                    user=self.user,
                    port=self.port,
                    config=config,
                    gateway=self.gateway,
                    connect_kwargs=connect_kwargs,
                )

        else:
            self._connection = self._get_single_connection(
                host=self.host,
                user=self.user,
                port=self.port,
                config=self.config,
                gateway=self.gateway,
                connect_kwargs=self.connect_kwargs,
            )

    def _get_single_connection(
        self,
        host,
        user,
        port,
        config,
        gateway,
        connect_kwargs,
    ):
        """Helper method to generate a fabric Connection given standard parameters."""
        from jobflow_remote.config.base import ConnectionData

        if isinstance(gateway, ConnectionData):
            gateway = self._get_single_connection(
                host=gateway.host,
                user=gateway.user,
                port=gateway.port,
                config=None,
                gateway=gateway.gateway,
                connect_kwargs=gateway.get_connect_kwargs(),
            )

        return fabric.Connection(
            host=host,
            user=user,
            port=port,
            config=config,
            gateway=gateway,
            forward_agent=self.forward_agent,
            connect_timeout=self.connect_timeout,
            connect_kwargs=connect_kwargs,
            inline_ssh_env=self.inline_ssh_env,
        )

    def __eq__(self, other):
        if not isinstance(other, RemoteHost):
            return False
        return self.as_dict() == other.as_dict()

    @property
    def connection(self):
        return self._connection

    def execute(
        self,
        command: str | list[str],
        workdir: str | Path | None = None,
        timeout: int | None = None,
    ):
        """Execute the given command on the host.

        Parameters
        ----------
        command: str or list of str
            Command to execute, as a str or list of str.
        workdir: str or None
            path where the command will be executed.

        Returns
        -------
        stdout : str
            Standard output of the command
        stderr : str
            Standard error of the command
        exit_code : int
            Exit code of the command.
        """
        self._check_connected()

        if isinstance(command, (list, tuple)):
            command = " ".join(command)

        command = self.sanitize_command(command)

        # TODO: check if this works:
        if not workdir:
            workdir = "."

        workdir = Path(workdir)

        timeout = timeout or self.timeout_execute

        if self.shell_cmd:
            shell_cmd = self.shell_cmd
            if self.login_shell:
                shell_cmd += " -l "
            shell_cmd += " -c "
            remote_command = shell_cmd + shlex.quote(command)
        else:
            remote_command = command

        out = self._execute_remote_func(
            lambda host: host.connection.run,
            remote_command,
            hide=True,
            warn=True,
            timeout=timeout,
            _connection_workdir=workdir,
        )

        stdout = self.sanitize_output(out.stdout)
        stderr = self.sanitize_output(out.stderr)

        return stdout, stderr, out.exited

    def mkdir(
        self, directory: str | Path, recursive: bool = True, exist_ok: bool = True
    ) -> bool:
        """Create directory on the host."""
        directory = Path(directory)
        command = f"mkdir {'-p ' if recursive else ''}{str(directory)!r}"
        try:
            stdout, stderr, returncode = self.execute(command)
            if returncode != 0:
                logger.warning(
                    f"Error creating folder {directory}. stdout: {stdout}, stderr: {stderr}"
                )
            else:
                return returncode == 0
        except Exception:
            logger.warning(f"Error creating folder {directory}", exc_info=True)
        return False

    def write_text_file(self, filepath: str | Path, content: str) -> None:
        """Write content to a file on the host."""
        self._check_connected()

        f = io.StringIO(content)

        self._execute_remote_func(lambda host: host.connection.put, f, str(filepath))

    def connect(self) -> None:
        self.connection.open()
        if self.keepalive:
            # create all the nested connections for all the gateways.
            connection = self.connection
            while connection:
                if isinstance(connection, fabric.Connection):
                    connection.transport.set_keepalive(self.keepalive)
                    connection = connection.gateway
                else:
                    connection = None

    def close(self) -> bool:
        connection = self.connection
        all_closed = True
        while connection:
            try:
                if isinstance(connection, fabric.Connection):
                    connection.close()
                    connection = connection.gateway
                else:
                    connection = None
            except Exception:
                all_closed = False
        return all_closed

    @property
    def is_connected(self) -> bool:
        return self.connection.is_connected

    def put(self, src, dst) -> None:
        self._check_connected()

        self._execute_remote_func(lambda host: host.connection.put, src, dst)

    def get(self, src, dst) -> None:
        self._check_connected()

        self._execute_remote_func(lambda host: host.connection.get, src, dst)

    def copy(self, src, dst) -> None:
        cmd = ["cp", str(src), str(dst)]
        self.execute(cmd)

    def _execute_remote_func(
        self,
        remote_cmd: Callable,
        *args,
        _connection_workdir: str | Path | None = None,
        **kwargs,
    ):
        """
        Execute a remote command using the host, with the option to
        recreate the connection and retry the command in case of failure.

        Parameters
        ----------
        remote_cmd
            A function that takes the current RemoteHost instance as its argument
            and returns the actual function to execute with the given args/kwargs
        args
            Positional arguments to pass to the remote command
        _connection_workdir:
            a path passed to Connection.cd to be used as working directory for the
            command.
        kwargs
            Keyword arguments to pass to the remote command
        """

        # The function needs to be called in this way because if retry_on_closed_connection
        # is True and the first execution of the command fails a new instance self._connection
        # will be recreated. Since remote_cmd are typically methods of the Connection object,
        # this ensures that the method is called on the actual instance of the connection, and
        # not on a previous one.
        def execute_cmd():
            # create the function associated to the current connection
            with ExitStack() as stack:
                if _connection_workdir:
                    stack.enter_context(self.connection.cd(_connection_workdir))
                func = remote_cmd(self)
                return func(*args, **kwargs)

        if self.retry_on_closed_connection:
            try:
                return execute_cmd()
            except OSError as e:
                msg = getattr(e, "message", str(e))
                error: BaseException = e
                if "Socket is closed" not in msg:
                    raise
            except SSHException as e:
                error = e
                msg = getattr(e, "message", str(e))
                if "Server connection dropped" not in msg:
                    raise
            except EOFError as e:
                error = e
        else:
            return execute_cmd()

        # if the code gets here one of the errors that could be due to drop of the
        # connection occurred. Try to close and reopen the connection and retry
        # one more time
        # Call to traceback.format_exception compatible with python 3.9
        logger.warning(
            f"Error while trying to execute a command on host {self.host}:\n"
            f"{''.join(traceback.format_exception(type(error), error, error.__traceback__))}"
            "Probably due to the connection dropping. "
            "Will reopen the connection and retry."
        )
        try:
            self.connection.close()
        except Exception:
            logger.warning(
                "Error while closing the connection during a retry. "
                "Proceeding with the retry.",
                exc_info=True,
            )
        self._create_connection()
        self.connect()
        return execute_cmd()

    def listdir(self, path: str | Path):
        self._check_connected()

        try:
            return self._execute_remote_func(
                lambda host: host.connection.sftp().listdir, str(path)
            )
        except FileNotFoundError:
            return []

    def remove(self, path: str | Path) -> None:
        self._check_connected()

        self._execute_remote_func(lambda host: host.connection.sftp().remove, str(path))

    def rmtree(self, path: str | Path, raise_on_error: bool = False) -> bool:
        """Recursively delete a directory tree on a remote host.

        It is intended to remove an entire directory tree, including all files
        and subdirectories, on this remote host.

        Parameters
        ----------
        path : str or Path
            The path to the directory tree to be removed.

        raise_on_error : bool
            If set to `False` (default), errors will be ignored, and the method will
            attempt to continue removing remaining files and directories.
            Otherwise, any errors encountered during the removal process
            will raise an exception.

        Returns
        -------
        bool
            True if the directory tree was successfully removed, False otherwise.
        """
        stdout, stderr, exit_code = self.execute(f"rm -r {path}")
        if exit_code != 0:
            msg = f"Error while deleting folder {path}. stdout: {stdout}, stderr: {stderr}"
            if raise_on_error:
                raise RuntimeError(msg)
            warnings.warn(msg, stacklevel=2)
            return False
        return True

    def _check_connected(self) -> bool:
        """
        Helper method to determine if fabric consider the connection open and
        open it otherwise.

        Since many operations requiring connections happen in the runner,
        if the connection drops there are cases where the host may not be
        reconnected. To avoid this issue always try to reconnect automatically
        if the connection is not open.

        Returns
        -------
        True if the connection is open.
        """
        if not self.is_connected:
            # Note: raising here instead of reconnecting demonstrated to be a
            # problem for how the queue managers are handled in the Runner.
            self.connect()
        return True

    @property
    def interactive_login(self) -> bool:
        return self._interactive_login


def inter_handler(title, instructions, prompt_list):
    """
    Handler function for interactive prompts from the server.
    Used by Interactive AuthSource.
    """
    if title:
        print(title.strip())
    if instructions:
        print(instructions.strip())

    resp = []  # Initialize the response container

    # Walk the list of prompts that the server sent that we need to answer
    for pr in prompt_list:
        in_value = input(pr[0]) if pr[1] else getpass.getpass(pr[0])
        resp.append(in_value)

    return tuple(resp)  # Convert the response list to a tuple and return it


class Interactive(AuthSource):
    """
    Interactive AuthSource. Prompts the user for all the requests coming
    from the server.
    """

    def __init__(self, username) -> None:
        super().__init__(username=username)

    def __repr__(self) -> str:
        return super()._repr(user=self.username)  # type: ignore[misc]

    def authenticate(self, transport):
        return transport.auth_interactive(self.username, inter_handler)


class InteractiveAuthStrategy(OpenSSHAuthStrategy):
    """
    AuthStrategy based on OpenSSHAuthStrategy that tries to use public keys
    and then switches to an interactive approach forwarding the requests
    from the server.
    """

    def get_sources(self):
        # get_pubkeys from OpenSSHAuthStrategy
        # With the current implementation exceptions may be raised in case if a key
        # in ~/.ssh cannot be parsed.
        # In addition other error ("Oops, unhandled type 3 ('unimplemented')")
        # can lead to the procedure being stuck. Don't try the keys at the moment
        # InteractiveAuthStrategy works for password only
        # try:
        #     yield from self.get_pubkeys()
        # except Exception as e:
        #     logger.warning(
        #         "Error while trying the authentication with all the public keys "
        #         f"available: {getattr(e, 'message', str(e))}. This may be due to the "
        #         "format of one of the keys. Authentication will proceed with "
        #         "interactive prompts"
        #     )

        yield Interactive(username=self.username)



"""The Runner orchestrating the Jobs execution."""

from __future__ import annotations

import json
import logging
import math
import shutil
import signal
import time
import traceback
import uuid
from collections import defaultdict
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING

from monty.json import MontyDecoder
from monty.os import makedirs_p
from qtoolkit.core.data_objects import QState, SubmissionStatus
from rich.prompt import Confirm

from jobflow_remote import JobController
from jobflow_remote.config.base import (
    ConfigError,
    ExecutionConfig,
    LogLevel,
    Project,
    RunnerOptions,
    WorkerBase,
)
from jobflow_remote.config.manager import ConfigManager
from jobflow_remote.jobs.batch import RemoteBatchManager
from jobflow_remote.jobs.data import IN_FILENAME, OUT_FILENAME, RemoteError
from jobflow_remote.jobs.state import JobState
from jobflow_remote.remote.data import (
    get_job_path,
    get_local_data_path,
    get_remote_in_file,
    get_remote_store,
    get_remote_store_filenames,
    resolve_job_dict_args,
)
from jobflow_remote.remote.queue import ERR_FNAME, OUT_FNAME, QueueManager, set_name_out
from jobflow_remote.utils.data import suuid
from jobflow_remote.utils.log import initialize_runner_logger
from jobflow_remote.utils.remote import UnsafeDeletionError, safe_remove_job_files
from jobflow_remote.utils.schedule import SafeScheduler

if TYPE_CHECKING:
    from jobflow_remote.remote.host import BaseHost
    from jobflow_remote.utils.db import MongoLock

logger = logging.getLogger(__name__)


class Runner:
    """
    Object orchestrating the execution of all the Jobs.

    Advances the status of the Jobs, handles the communication with the workers
    and updates the queue and output databases.

    The main entry point is the `run` method. It is supposed to be executed
    by a daemon, but can also be run directly for testing purposes.
    It allows to run all the steps required to advance the Job's states or even
    a subset of them, to parallelize the different tasks.

    The runner instantiates a pool of workers and hosts given in the project
    definition. A single connection will be opened if multiple workers share
    the same host.

    The Runner schedules the execution of the specific tasks at regular intervals
    and relies on objects like QueueManager, BaseHost and JobController to
    interact with workers and databases.
    """

    def __init__(
        self,
        project_name: str | None = None,
        log_level: LogLevel | None = None,
        runner_id: str | None = None,
        connect_interactive: bool = False,
    ) -> None:
        """
        Parameters
        ----------
        project_name
            Name of the project. Used to retrieve all the configurations required
            to execute the runner.
        log_level
            Logging level of the Runner.
        runner_id
            A unique identifier for the Runner process. Used to identify the
            runner process in logging and in the DB locks.
            If None a uuid will be generated.
        connect_interactive:
            If True during initialization will open connections to the hosts
            marked as interactive.
        """
        self.stop_signal = False
        self.runner_id: str = runner_id or str(uuid.uuid4())
        self.config_manager: ConfigManager = ConfigManager()
        self.project_name = project_name
        self.project: Project = self.config_manager.get_project(project_name)
        self.job_controller: JobController = JobController.from_project(self.project)
        self.workers: dict[str, WorkerBase] = self.project.workers
        # Build the dictionary of hosts. The reference is the worker name.
        # If two hosts match, use the same instance
        self.hosts: dict[str, BaseHost] = {}
        for wname, w in self.workers.items():
            new_host = w.get_host()
            for host in self.hosts.values():
                if new_host == host:
                    self.hosts[wname] = host
                    break
            else:
                self.hosts[wname] = new_host
        self.limited_workers = {
            name: {"max": w.max_jobs, "current": 0}
            for name, w in self.workers.items()
            if w.max_jobs and not w.batch
        }
        self.batch_workers = {}
        for wname, w in self.workers.items():
            if w.batch is not None:
                self.batch_workers[wname] = RemoteBatchManager(
                    self.hosts[wname], w.batch.jobs_handle_dir
                )
        self.queue_managers: dict = {}
        log_level = log_level if log_level is not None else self.project.log_level
        initialize_runner_logger(
            log_folder=self.project.log_dir,
            level=log_level.to_logging(),
            runner_id=self.runner_id,
        )
        # TODO it could be better to create a pool of stores that are connected
        # How to deal with cases where the connection gets closed?
        # how to deal with file based stores?
        self.jobstore = self.project.get_jobstore()

        if connect_interactive:
            for host_name, host in self.hosts.items():
                if (
                    host.interactive_login
                    and not host.is_connected
                    and Confirm.ask(
                        f"Do you want to open the connection for the host of the {host_name} worker?"
                    )
                ):
                    host.connect()

    @property
    def runner_options(self) -> RunnerOptions:
        """The Runner options defined in the project."""
        return self.project.runner

    def handle_signal(self, signum, frame) -> None:
        """
        Handle the SIGTERM signal in the Runner.
        Sets a variable that will stop the Runner loop.
        """
        logger.info(f"Received signal: {signum}")
        self.stop_signal = True

    def get_worker(self, worker_name: str) -> WorkerBase:
        """
        Get the worker from the pool of workers instantiated by the Runner.

        Parameters
        ----------
        worker_name
            The name of the worker.

        Returns
        -------
            An instance of the corresponding worker.
        """
        if worker_name not in self.workers:
            raise ConfigError(
                f"No worker {worker_name} is defined in project {self.project_name}"
            )
        return self.workers[worker_name]

    def get_host(self, worker_name: str) -> BaseHost:
        """
        Get the host associated to a worker from the pool of hosts instantiated
        by the Runner.

        Parameters
        ----------
        worker_name
            The name of the worker.

        Returns
        -------
            An instance of the Host associated to the worker.
        """
        host = self.hosts[worker_name]
        if not host.is_connected:
            host.connect()
        return host

    def get_queue_manager(self, worker_name: str) -> QueueManager:
        """
        Get an instance of the queue manager associated to a worker, based on its host.

        Parameters
        ----------
        worker_name
            The name of the worker.

        Returns
        -------
             An instance of the QueueManager associated to the worker.
        """
        if worker_name not in self.queue_managers:
            worker = self.get_worker(worker_name)
            self.queue_managers[worker_name] = QueueManager(
                worker.get_scheduler_io(), self.get_host(worker_name)
            )
        return self.queue_managers[worker_name]

    def run(
        self,
        transfer: bool = True,
        complete: bool = True,
        queue: bool = True,
        checkout: bool = True,
        ticks: int | None = None,
    ) -> None:
        """
        Start the runner.

        Which actions are being performed can be tuned by the arguments.

        Parameters
        ----------
        transfer
            If True actions related to file transfer are performed by the runner.
        complete
            If True Job completion is performed by the runner.
        queue
            If True interactions with the queue manager are handled by the Runner.
        checkout
            If True the checkout of Jobs is performed by the Runner.
        ticks
            If provided, the Runner will run for this number of ticks before exiting.
        """
        signal.signal(signal.SIGTERM, self.handle_signal)

        states = []
        if transfer:
            states.append(JobState.CHECKED_OUT.value)
            states.append(JobState.TERMINATED.value)
        if complete:
            states.append(JobState.DOWNLOADED.value)
        if queue:
            states.append(JobState.UPLOADED.value)

        logger.info(
            f"Runner run options: transfer: {transfer} complete: {complete} queue: {queue} checkout: {checkout}"
        )

        scheduler = SafeScheduler(seconds_after_failure=120)

        # run a first call for each case, since schedule will wait for the delay
        # to make the first execution.
        if checkout:
            try:
                self.checkout()
            except Exception:
                logger.exception("Error during initial checkout")
            scheduler.every(self.runner_options.delay_checkout).seconds.do(
                self.checkout
            )

        if transfer or queue or complete:
            try:
                self.advance_state(states)
            except Exception:
                logger.exception("Error during initial advance_state")
            scheduler.every(self.runner_options.delay_advance_status).seconds.do(
                self.advance_state, states=states
            )

        if queue:
            try:
                self.check_run_status()
            except Exception:
                logger.exception("Error during initial check_run_status")
            scheduler.every(self.runner_options.delay_check_run_status).seconds.do(
                self.check_run_status
            )
            # Limited workers will only affect the process interacting with the queue
            # manager. When a job is submitted or terminated the count in the
            # limited_workers can be directly updated, since by construction only one
            # process will take care of the queue state.
            # The refresh can be run on a relatively high delay since it should only
            # account for actions from the user (e.g. rerun, cancel), that can alter
            # the number of submitted/running jobs.
            if self.limited_workers:
                try:
                    self.refresh_num_current_jobs()
                except Exception:
                    logger.exception("Error during initial refresh_num_current_jobs")
                scheduler.every(self.runner_options.delay_refresh_limited).seconds.do(
                    self.refresh_num_current_jobs
                )
            if self.batch_workers:
                try:
                    self.update_batch_jobs()
                except Exception:
                    logger.exception("Error during initial update_batch_jobs")
                scheduler.every(self.runner_options.delay_update_batch).seconds.do(
                    self.update_batch_jobs
                )

        # all the processes will ping the running_runner document to signal
        # that at least one is still active.
        scheduler.every(self.project.runner.delay_ping_db).seconds.do(
            self.ping_running_runner
        )

        ticks_remaining: int | bool = True
        if ticks is not None:
            ticks_remaining = ticks

        while ticks_remaining:
            if self.stop_signal:
                logger.info("stopping due to sigterm")
                break
            scheduler.run_pending()
            time.sleep(1)

            if ticks is not None:
                ticks_remaining -= 1

    def run_all_jobs(
        self,
        max_seconds: int | None = None,
    ) -> None:
        """
        Use the runner to run all the jobs in the DB.
        Mainly used for testing.
        """
        states = [
            JobState.CHECKED_OUT.value,
            JobState.TERMINATED.value,
            JobState.DOWNLOADED.value,
            JobState.UPLOADED.value,
        ]

        scheduler = SafeScheduler(seconds_after_failure=120)

        t0 = time.time()
        # run a first call for each case, since schedule will wait for the delay
        # to make the first execution.
        self.checkout()
        scheduler.every(self.runner_options.delay_checkout).seconds.do(self.checkout)

        self.advance_state(states)
        scheduler.every(self.runner_options.delay_advance_status).seconds.do(
            self.advance_state, states=states
        )

        self.check_run_status()
        scheduler.every(self.runner_options.delay_check_run_status).seconds.do(
            self.check_run_status
        )

        # Limited workers will only affect the process interacting with the queue
        # manager. When a job is submitted or terminated the count in the
        # limited_workers can be directly updated, since by construction only one
        # process will take care of the queue state.
        # The refresh can be run on a relatively high delay since it should only
        # account for actions from the user (e.g. rerun, cancel), that can alter
        # the number of submitted/running jobs.
        if self.limited_workers:
            self.refresh_num_current_jobs()
            scheduler.every(self.runner_options.delay_refresh_limited).seconds.do(
                self.refresh_num_current_jobs
            )
        if self.batch_workers:
            self.update_batch_jobs()
            scheduler.every(self.runner_options.delay_update_batch).seconds.do(
                self.update_batch_jobs
            )

        running_states = [
            JobState.READY.value,
            JobState.CHECKED_OUT.value,
            JobState.TERMINATED.value,
            JobState.DOWNLOADED.value,
            JobState.UPLOADED.value,
            JobState.SUBMITTED.value,
            JobState.RUNNING.value,
            JobState.BATCH_RUNNING.value,
            JobState.BATCH_SUBMITTED.value,
        ]
        query = {"state": {"$in": running_states}}
        jobs_available = 1
        while jobs_available:
            scheduler.run_pending()
            time.sleep(0.2)
            jobs_available = self.job_controller.count_jobs(query=query)
            if max_seconds and time.time() - t0 > max_seconds:
                raise RuntimeError(
                    "Could not execute all the jobs within the selected amount of time"
                )

    def run_one_job(
        self,
        db_id: str | None = None,
        job_id: tuple[str, int] | None = None,
        max_seconds: int | None = None,
        raise_at_timeout: bool = True,
        target_state: JobState | None = None,
    ) -> bool:
        """
        Use the runner to run a single Job until it reaches a terminal state.
        The job should be in the READY state and there should be no
        Mainly used for testing.
        """
        states = [
            JobState.CHECKED_OUT.value,
            JobState.TERMINATED.value,
            JobState.DOWNLOADED.value,
            JobState.UPLOADED.value,
        ]

        scheduler = SafeScheduler(seconds_after_failure=120)

        t0 = time.time()
        query: dict = {}
        if db_id:
            query["db_id"] = db_id
        if job_id:
            query["uuid"] = job_id[0]
            query["index"] = job_id[1]
        job_data = self.job_controller.checkout_job(query=query)
        if not job_data:
            if not db_id and not job_id:
                return False
            if not db_id:
                job_data = job_id
            else:
                j_info = self.job_controller.get_job_info(db_id=db_id)
                job_data = (j_info.uuid, j_info.index)

        filters = {"uuid": job_data[0], "index": job_data[1]}
        self.advance_state(states)
        scheduler.every(self.runner_options.delay_advance_status).seconds.do(
            self.advance_state,
            states=states,
            filter=filters,
        )

        self.check_run_status()
        scheduler.every(self.runner_options.delay_check_run_status).seconds.do(
            self.check_run_status, filter=filters
        )

        running_states = [
            JobState.READY.value,
            JobState.CHECKED_OUT.value,
            JobState.TERMINATED.value,
            JobState.DOWNLOADED.value,
            JobState.UPLOADED.value,
            JobState.SUBMITTED.value,
            JobState.RUNNING.value,
        ]

        while True:
            scheduler.run_pending()
            time.sleep(0.2)
            job_info = self.job_controller.get_job_info(
                job_id=job_data[0], job_index=job_data[1]
            )
            if target_state and job_info.state == target_state:
                return True
            if job_info.state.value not in running_states:
                # if the target state is defined and the code got
                # here, it means it missed the target state, so
                # the target was not achieved.
                if target_state:
                    raise RuntimeError(
                        f"The target state {target_state.value} was not achieved. "
                        f"Final state: {job_info.state.value}"
                    )
                return True
            if max_seconds and time.time() - t0 > max_seconds:
                if raise_at_timeout:
                    raise RuntimeError(
                        "Could not execute the job within the selected amount of time"
                    )
                return False

    def _get_limited_worker_query(self, states: list[str]) -> dict | None:
        """
        Generate the query to be used for fetching Jobs for workers with limited
        number of Jobs allowed.

        Parameters
        ----------
        states
            The states to be used in the query.

        Returns
        -------
            A dictionary with the query.
        """
        states = [s for s in states if s != JobState.UPLOADED.value]

        available_workers = [w for w in self.workers if w not in self.limited_workers]
        for worker, status in self.limited_workers.items():
            if status["current"] < status["max"]:
                available_workers.append(worker)

        states_query = {"state": {"$in": states}}
        uploaded_query = {
            "state": JobState.UPLOADED.value,
            "worker": {"$in": available_workers},
        }

        if states and available_workers:
            return {"$or": [states_query, uploaded_query]}
        if states:
            return states_query
        if available_workers:
            return uploaded_query

        return None

    def advance_state(self, states: list[str], filter: dict | None = None) -> None:  # noqa: A002
        """
        Acquire the lock and advance the state of a single job.

        Parameters
        ----------
        states
            The state of the Jobs that can be queried.
        """
        states_methods = {
            JobState.CHECKED_OUT: self.upload,
            JobState.UPLOADED: self.submit,
            JobState.TERMINATED: self.download,
            JobState.DOWNLOADED: self.complete_job,
        }

        while True:
            # handle the case of workers with limited number of jobs
            if self.limited_workers and JobState.UPLOADED.value in states:
                query = self._get_limited_worker_query(states=states)
                if not query:
                    return
            else:
                query = {"state": {"$in": states}}
            if filter:
                query.update(filter)

            with self.job_controller.lock_job_for_update(
                query=query,
                max_step_attempts=self.runner_options.max_step_attempts,
                delta_retry=self.runner_options.delta_retry,
            ) as lock:
                doc = lock.locked_document
                if not doc:
                    return

                state = JobState(doc["state"])

                states_methods[state](lock)

    def upload(self, lock: MongoLock) -> None:
        """
        Upload files for a locked Job in the CHECKED_OUT state.
        If successful set the state to UPLOADED.

        Parameters
        ----------
        lock
            The MongoLock with the locked Job document.
        """
        doc = lock.locked_document
        if doc is None:
            raise RuntimeError("No document found in the lock.")

        db_id = doc["db_id"]
        logger.debug(f"upload db_id: {db_id}")

        job_dict = doc["job"]

        worker = self.get_worker(doc["worker"])
        host = self.get_host(doc["worker"])

        # if prerun_cleanup is specified (job was likely rerun) delete the folder before
        # reuploading the data
        run_dir = doc["run_dir"]
        if doc.get("remote", {}).get("prerun_cleanup", False):
            try:
                deleted = safe_remove_job_files(
                    host=host, run_dir=run_dir, raise_on_error=True
                )
                # Log as debug. If deletion failed because of an error it will raise.
                not_string = "" if deleted else "not "
                logger.debug(
                    f"Folder for job {db_id} ({run_dir}) was {not_string}deleted"
                )
            except UnsafeDeletionError as e:
                raise RemoteError(
                    f"Error while performing cleanup of the run_dir folder for job {db_id}: {run_dir}",
                    no_retry=True,
                ) from e
            except Exception as e:
                raise RemoteError(
                    f"Error while performing cleanup of the run_dir folder for job {db_id}: {run_dir}",
                    no_retry=False,
                ) from e

        store = self.jobstore
        # TODO would it be better/feasible to keep a pool of the required
        # Stores already connected, to avoid opening and closing them?
        store.connect()
        try:
            resolve_job_dict_args(job_dict, store)
        finally:
            try:
                store.close()
            except Exception:
                logging.exception(f"error while closing the store {store}")

        remote_path = get_job_path(job_dict["uuid"], job_dict["index"], worker.work_dir)

        # Set the value of the original store for dynamical workflow. Usually it
        # will be None don't add the serializer, at this stage the default_orjson
        # serializer could undergo refactoring and this could break deserialization
        # of older FWs. It is set in the FireTask at runtime.
        remote_store = get_remote_store(
            store=store, work_dir=remote_path, config_dict=self.project.remote_jobstore
        )

        created = host.mkdir(remote_path)
        if not created:
            err_msg = (
                f"Could not create remote directory {remote_path} for db_id {db_id}"
            )
            logger.error(err_msg)
            raise RemoteError(err_msg, no_retry=False)

        serialized_input = get_remote_in_file(job_dict, remote_store, doc)

        path_file = Path(remote_path, IN_FILENAME)
        host.put(serialized_input, str(path_file))

        set_output = {
            "$set": {
                "run_dir": remote_path,
                "state": JobState.UPLOADED.value,
                "remote.prerun_cleanup": False,
            }
        }
        lock.update_on_release = set_output

    def submit(self, lock: MongoLock) -> None:
        """
        Submit to the queue for a locked Job in the UPLOADED state.
        If successful set the state to SUBMITTED.

        Parameters
        ----------
        lock
            The MongoLock with the locked Job document.
        """
        doc = lock.locked_document
        if doc is None:
            raise RuntimeError("No document found in the lock.")

        logger.debug(f"submit db_id: {doc['db_id']}")

        job_dict = doc["job"]

        worker_name = doc["worker"]
        worker = self.get_worker(worker_name)

        remote_path = Path(doc["run_dir"])

        execution_cmd = worker.execution_cmd or "jf -fe execution run {}"
        script_commands = [execution_cmd.format(remote_path)]

        queue_manager = self.get_queue_manager(worker_name)
        qout_fpath = remote_path / OUT_FNAME
        qerr_fpath = remote_path / ERR_FNAME

        exec_config = doc["exec_config"]
        if isinstance(exec_config, str):
            exec_config = self.config_manager.get_exec_config(
                exec_config_name=exec_config, project_name=self.project_name
            )
        elif isinstance(exec_config, dict):
            exec_config = ExecutionConfig.parse_obj(exec_config)

        # define an empty default if it is not set
        exec_config = exec_config or ExecutionConfig()

        if worker_name in self.batch_workers:
            resources: dict = {}

            set_name_out(
                resources, job_dict["name"], out_fpath=qout_fpath, err_fpath=qerr_fpath
            )
            shell_manager = queue_manager.get_shell_manager()
            shell_manager.write_submission_script(
                commands=script_commands,
                pre_run=exec_config.pre_run,
                post_run=exec_config.post_run,
                options=resources,
                export=exec_config.export,
                modules=exec_config.modules,
                work_dir=remote_path,
                create_submit_dir=False,
            )

            self.batch_workers[worker_name].submit_job(
                job_id=doc["uuid"], index=doc["index"]
            )
            lock.update_on_release = {
                "$set": {
                    "state": JobState.BATCH_SUBMITTED.value,
                }
            }
        else:
            # decode in case it contains a QResources. It was not deserialized before.
            resources = (
                MontyDecoder().process_decoded(doc["resources"])
                or worker.resources
                or {}
            )
            set_name_out(
                resources, job_dict["name"], out_fpath=qout_fpath, err_fpath=qerr_fpath
            )

            pre_run = worker.pre_run or ""
            if exec_config.pre_run:
                pre_run += "\n" + exec_config.pre_run
            post_run = worker.post_run or ""
            if exec_config.post_run:
                post_run += "\n" + exec_config.post_run

            submit_result = queue_manager.submit(
                commands=script_commands,
                pre_run=pre_run,
                post_run=post_run,
                options=resources,
                export=exec_config.export,
                modules=exec_config.modules,
                work_dir=remote_path,
                create_submit_dir=False,
            )

            if submit_result.status == SubmissionStatus.FAILED:
                err_msg = f"submission failed. {submit_result!r}"
                raise RemoteError(err_msg, no_retry=False)
            if submit_result.status == SubmissionStatus.JOB_ID_UNKNOWN:
                err_msg = (
                    "submission succeeded but ID not known. Job may be running "
                    f"but status cannot be checked. {submit_result!r}"
                )
                raise RemoteError(err_msg, no_retry=True)
            if submit_result.status == SubmissionStatus.SUCCESSFUL:
                lock.update_on_release = {
                    "$set": {
                        "remote.process_id": str(submit_result.job_id),
                        "state": JobState.SUBMITTED.value,
                    }
                }
                if worker_name in self.limited_workers:
                    self.limited_workers[worker_name]["current"] += 1
            else:
                raise RemoteError(
                    f"unhandled submission status {submit_result.status}", no_retry=True
                )

    def download(self, lock) -> None:
        """
        Download the final files for a locked Job in the TERMINATED state.
        If successful set the state to DOWNLOADED.

        Parameters
        ----------
        lock
            The MongoLock with the locked Job document.
        """
        doc = lock.locked_document
        logger.debug(f"download db_id: {doc['db_id']}")

        # job_doc = JobDoc(**doc)
        job_dict = doc["job"]

        # If the worker is local do not copy the files in the temporary folder
        # It should not arrive to this point, since it should go directly
        # from SUBMITTED/RUNNING to DOWNLOADED in case of local worker
        worker = self.get_worker(doc["worker"])
        if not worker.is_local:
            host = self.get_host(doc["worker"])
            store = self.jobstore

            remote_path = doc["run_dir"]
            local_path = get_local_data_path(
                project=self.project,
                worker=worker,
                job_id=doc["uuid"],
                index=doc["index"],
                run_dir=remote_path,
            )

            makedirs_p(local_path)

            def download_file(fname: str, mandatory: bool):
                # in principle fabric should work by just passing the
                # destination folder, but it fails
                remote_file_path = str(Path(remote_path, fname))
                local_file_path = Path(local_path, fname)
                try:
                    host.get(remote_file_path, str(local_file_path))
                except FileNotFoundError as exc:
                    # fabric may still create an empty local file even if the remote
                    # file does not exist. Remove it to avoid errors when checking
                    # the file existence
                    local_file_path.unlink(missing_ok=True)
                    err_msg = f"file {remote_file_path} for job {job_dict['uuid']} does not exist"
                    if mandatory:
                        logger.exception(err_msg)
                        # if files are missing it should not retry
                        raise RemoteError(err_msg, no_retry=True) from exc

                    err_msg += ". Allow continuing to the next state"
                    logger.warning(err_msg)

            # download the queue files first, so if an error is triggered
            # afterwards they can be inserted in the DB
            for fn in ("queue.out", "queue.err"):
                download_file(fn, mandatory=False)

            # only the output file is mandatory. If the others are missing
            # it will be dealt with by the complete. The output file may contain
            # an error and the fact that they are missing could be expected.
            # The analysis of the output file is left to the completion procedure.
            download_file(OUT_FILENAME, mandatory=True)

            for fn in get_remote_store_filenames(
                store, config_dict=self.project.remote_jobstore
            ):
                download_file(fn, mandatory=False)

        lock.update_on_release = {"$set": {"state": JobState.DOWNLOADED.value}}

    def complete_job(self, lock) -> None:
        """
        Complete a locked Job in the DOWNLOADED state.
        If successful set the state to COMPLETED, otherwise to FAILED.

        Parameters
        ----------
        lock
            The MongoLock with the locked Job document.
        """
        doc = lock.locked_document
        logger.debug(f"complete job db_id: {doc['db_id']}")

        # if the worker is local the files were not copied to the temporary
        # folder, but the files could be directly accessed
        worker = self.get_worker(doc["worker"])
        local_path = get_local_data_path(
            project=self.project,
            worker=worker,
            job_id=doc["uuid"],
            index=doc["index"],
            run_dir=doc["run_dir"],
        )

        try:
            store = self.jobstore
            completed = self.job_controller.complete_job(doc, local_path, store)

        except json.JSONDecodeError as exc:
            # if an empty file is copied this error can appear, do not retry
            err_msg = traceback.format_exc()
            raise RemoteError(err_msg, no_retry=True) from exc

        # remove local folder with downloaded files if successfully completed
        if completed and self.runner_options.delete_tmp_folder and not worker.is_local:
            shutil.rmtree(local_path, ignore_errors=True)

        if not completed:
            err_msg = "the parsed output does not contain the required information to complete the job"
            raise RemoteError(err_msg, no_retry=True)

    def check_run_status(self, filter: dict | None = None) -> None:  # noqa: A002
        """
        Check the status of all the jobs submitted to a queue.

        If Jobs started update their state from SUBMITTED to RUNNING.
        If Jobs terminated set their state to TERMINATED if running on a remote
        host. If on a local host set them directly to DOWNLOADED.
        """
        logger.debug("check_run_status")
        # check for jobs that could have changed state
        workers_ids_docs: dict = defaultdict(dict)
        db_filter = {
            "state": {"$in": [JobState.SUBMITTED.value, JobState.RUNNING.value]},
            "lock_id": None,
            "remote.retry_time_limit": {"$not": {"$gt": datetime.utcnow()}},
        }
        if filter:
            db_filter.update(filter)
        projection = [
            "db_id",
            "uuid",
            "index",
            "remote",
            "worker",
            "state",
        ]
        for doc in self.job_controller.get_jobs(db_filter, projection):
            worker_name = doc["worker"]
            remote_doc = doc["remote"]
            workers_ids_docs[worker_name][remote_doc["process_id"]] = doc

        for worker_name, ids_docs in workers_ids_docs.items():
            error = None
            if not ids_docs:
                continue

            worker = self.get_worker(worker_name)

            qjobs_dict = {}
            try:
                ids_list = list(ids_docs)
                queue = self.get_queue_manager(worker_name)
                qjobs = queue.get_jobs_list(
                    jobs=ids_list, user=worker.scheduler_username
                )
                qjobs_dict = {
                    qjob.job_id: qjob for qjob in qjobs if qjob.job_id in ids_list
                }
            except Exception:
                logger.warning(
                    f"error trying to get jobs list for worker: {worker_name}",
                    exc_info=True,
                )
                error = traceback.format_exc()

            for doc_id, doc in ids_docs.items():
                # TODO if failed should maybe be handled differently?
                remote_doc = doc["remote"]
                qjob = qjobs_dict.get(doc_id)
                qstate = qjob.state if qjob else None
                next_state = None
                start_time = None
                next_step_delay = None
                if (
                    qstate == QState.RUNNING
                    and doc["state"] == JobState.SUBMITTED.value
                ):
                    next_state = JobState.RUNNING
                    start_time = datetime.utcnow()
                    logger.debug(
                        f"remote job with id {remote_doc['process_id']} is running"
                    )
                elif qstate in [None, QState.DONE, QState.FAILED]:
                    # if the worker is local go directly to DOWNLOADED, as files
                    # are not copied locally
                    if not worker.is_local:
                        next_state = JobState.TERMINATED
                    else:
                        next_state = JobState.DOWNLOADED
                    # the delay is applied if the job is finished on the worker
                    next_step_delay = worker.delay_download
                    logger.debug(
                        f"terminated remote job with id {remote_doc['process_id']}"
                    )
                elif not error and remote_doc["step_attempts"] > 0:
                    # reset the step attempts if succeeding in case there was
                    # an error earlier. Setting the state to the same as the
                    # current triggers the update that cleans the state
                    next_state = JobState(doc["state"])

                # the document needs to be updated only in case of error or if a
                # next state has been set.
                # Only update if the state did not change in the meanwhile
                if next_state or error:
                    lock_filter = {
                        "uuid": doc["uuid"],
                        "index": doc["index"],
                        "state": doc["state"],
                    }

                    with self.job_controller.lock_job_for_update(
                        query=lock_filter,
                        max_step_attempts=self.runner_options.max_step_attempts,
                        delta_retry=self.runner_options.delta_retry,
                        next_step_delay=next_step_delay,
                    ) as lock:
                        if lock.locked_document:
                            if error:
                                raise RemoteError(error, no_retry=False)
                            set_output = {
                                "$set": {
                                    "remote.queue_state": (
                                        qstate.value if qstate else None
                                    ),
                                    "state": next_state.value,
                                }
                            }
                            if start_time:
                                set_output["$set"]["start_time"] = start_time
                            lock.update_on_release = set_output
                    # decrease the amount of jobs running if it is a limited worker
                    if (
                        next_state in (JobState.TERMINATED, JobState.DOWNLOADED)
                        and worker_name in self.limited_workers
                    ):
                        self.limited_workers[doc["worker"]]["current"] -= 1

    def checkout(self) -> None:
        """Checkout READY Jobs."""
        logger.debug("checkout jobs")
        n_checked_out = 0
        while True:
            try:
                reserved = self.job_controller.checkout_job()
                if not reserved:
                    break
            except Exception:
                logger.exception("Error while checking out jobs")
                break

            n_checked_out += 1

        logger.debug(f"checked out {n_checked_out} jobs")

    def refresh_num_current_jobs(self) -> None:
        """
        Update the number of jobs currently running for worker with limited
        number of Jobs.
        """
        for name, state in self.limited_workers.items():
            query = {
                "state": {"$in": [JobState.SUBMITTED.value, JobState.RUNNING.value]},
                "worker": name,
            }
            state["current"] = self.job_controller.count_jobs(query)

    def update_batch_jobs(self) -> None:
        """
        Update the status of batch jobs.

        Includes submitting to the remote queue, checking the status of
        running jobs in the queue and handle the files with the Jobs information
        about their status.
        """
        logger.debug("update batch jobs")
        for worker_name, batch_manager in self.batch_workers.items():
            worker = self.get_worker(worker_name)
            # first check the processes that are running from the folder
            # and set them to running if needed
            running_jobs = []
            try:
                running_jobs = batch_manager.get_running()
            except Exception:
                logger.warning(
                    f"error trying to get the list of batch running jobs for worker: {worker_name}",
                    exc_info=True,
                )

            for job_id, job_index, process_running_uuid in running_jobs:
                lock_filter = {
                    "uuid": job_id,
                    "index": job_index,
                    "state": JobState.BATCH_SUBMITTED.value,
                }
                with self.job_controller.lock_job_for_update(
                    query=lock_filter,
                    max_step_attempts=self.runner_options.max_step_attempts,
                    delta_retry=self.runner_options.delta_retry,
                ) as lock:
                    if lock.locked_document:
                        set_output = {
                            "$set": {
                                "state": JobState.BATCH_RUNNING.value,
                                "start_time": datetime.utcnow(),
                                "remote.process_id": process_running_uuid,
                            }
                        }
                        lock.update_on_release = set_output

            # Check the processes that should be running on the remote queue
            # and update the state in the DB if something changed
            batch_processes_data = self.job_controller.get_batch_processes(
                worker_name
            ).get(worker_name, {})
            processes = list(batch_processes_data)
            queue_manager = self.get_queue_manager(worker_name)
            if processes:
                stopped_processes = set()
                running_processes = set()
                try:
                    qjobs = queue_manager.get_jobs_list(
                        jobs=processes, user=worker.scheduler_username
                    )
                    running_processes = {
                        qjob.job_id for qjob in qjobs if qjob.job_id in processes
                    }
                    stopped_processes = set(processes) - running_processes
                except Exception:
                    logger.warning(
                        f"error trying to get the list of batch processes for worker: {worker_name}",
                        exc_info=True,
                    )

                for pid in stopped_processes:
                    self.job_controller.remove_batch_process(pid, worker_name)
                    # check if there are jobs that were in the running folder of a
                    # process that finished and set them to remote error
                    for job_id, job_index, process_running_uuid in running_jobs:
                        if batch_processes_data[pid] == process_running_uuid:
                            lock_filter = {
                                "uuid": job_id,
                                "index": job_index,
                                "state": {
                                    "$in": (
                                        JobState.BATCH_SUBMITTED.value,
                                        JobState.BATCH_RUNNING.value,
                                    )
                                },
                            }
                            with self.job_controller.lock_job_for_update(
                                query=lock_filter,
                                max_step_attempts=self.runner_options.max_step_attempts,
                                delta_retry=self.runner_options.delta_retry,
                            ) as lock:
                                if lock.locked_document:
                                    err_msg = f"The batch process that was running the job (process_id: {pid}, uuid: {process_running_uuid} was likely killed before terminating the job execution"
                                    raise RemoteError(err_msg, no_retry=True)

                processes = list(running_processes)

            # check that enough processes are submitted and submit the required
            # amount to reach max_jobs, if needed.

            dict_n_jobs = self.job_controller.count_jobs_states(
                [JobState.BATCH_SUBMITTED, JobState.BATCH_RUNNING]
            )
            n_jobs_submitted = dict_n_jobs[JobState.BATCH_SUBMITTED]
            n_jobs_running = dict_n_jobs[JobState.BATCH_RUNNING]

            n_processes = len(processes)
            n_parallel = worker.batch.parallel_jobs or 1
            # TODO for n_parallel > 1 here a new job is submitted to the queue even if only
            # one jobflow Job is available to run. This may result in the submitted job
            # be partially empty. Would it be better to submit only if more jobs need to be run?
            # (in that case a single job may remain dangling in the queue)

            # The purpose of dealing with running and submitted jobs separately is to avoid
            # submitting processes if a job remains stuck in a SBATCH_RUNNING state.
            # In principle it should not happen but if happening it will keep submitting
            # batch processes to the queue.
            available_jobs = max(
                n_jobs_submitted - max((n_processes * n_parallel) - n_jobs_running, 0),
                0,
            )
            n_processes_to_submit = min(
                max(worker.max_jobs - n_processes, 0),
                math.ceil(available_jobs / n_parallel),
            )
            logger.debug(
                f"submitting {n_processes_to_submit} batch jobs for worker {worker_name}"
            )
            for _ in range(n_processes_to_submit):
                resources = worker.resources or {}
                process_running_uuid = suuid()
                remote_path = Path(
                    get_job_path(process_running_uuid, None, worker.batch.work_dir)
                )
                qout_fpath = remote_path / OUT_FNAME
                qerr_fpath = remote_path / ERR_FNAME
                set_name_out(
                    resources,
                    f"batch_{process_running_uuid}",
                    out_fpath=qout_fpath,
                    err_fpath=qerr_fpath,
                )

                # note that here the worker.work_dir needs to be passed,
                # not the worker.batch.work_dir
                command = f"jf -fe execution run-batch {worker.work_dir} {worker.batch.jobs_handle_dir} {process_running_uuid}"
                if worker.batch.max_jobs_per_batch:
                    command += f" -mj {worker.batch.max_jobs_per_batch}"
                if worker.batch.max_time:
                    command += f" -mt {worker.batch.max_time}"
                if worker.batch.max_wait:
                    command += f" -mw {worker.batch.max_wait}"
                if worker.batch.parallel_jobs:
                    command += f" -pj {worker.batch.parallel_jobs}"

                submit_result = queue_manager.submit(
                    commands=[command],
                    pre_run=worker.pre_run,
                    post_run=worker.post_run,
                    options=resources,
                    work_dir=remote_path,
                    create_submit_dir=True,
                )

                if submit_result.status == SubmissionStatus.FAILED:
                    logger.error(f"submission failed. {submit_result!r}")
                elif submit_result.status == SubmissionStatus.JOB_ID_UNKNOWN:
                    logger.error(
                        f"submission succeeded but ID not known. Job may be running but status cannot be checked. {submit_result!r}"
                    )

                elif submit_result.status == SubmissionStatus.SUCCESSFUL:
                    logger.debug(
                        f"Batch job submitted to worker {worker_name} in folder {remote_path}. Queue id: {submit_result.job_id}"
                    )
                    self.job_controller.add_batch_process(
                        submit_result.job_id, process_running_uuid, worker_name
                    )
                else:
                    logger.error(f"unhandled submission status {submit_result.status}")

            # check for jobs that have terminated in the batch runner and
            # update the DB state accordingly
            terminated_jobs = []
            try:
                terminated_jobs = batch_manager.get_terminated()
            except Exception:
                logger.warning(
                    f"error trying to get the list of terminated batch jobs for worker: {worker_name}",
                    exc_info=True,
                )
            for job_id, job_index, process_running_uuid in terminated_jobs:
                lock_filter = {
                    "uuid": job_id,
                    "index": job_index,
                    "state": {
                        "$in": (
                            JobState.BATCH_SUBMITTED.value,
                            JobState.BATCH_RUNNING.value,
                        )
                    },
                }
                with self.job_controller.lock_job_for_update(
                    query=lock_filter,
                    max_step_attempts=self.runner_options.max_step_attempts,
                    delta_retry=self.runner_options.delta_retry,
                ) as lock:
                    if lock.locked_document:
                        if not worker.is_local:
                            next_state = JobState.TERMINATED
                        else:
                            next_state = JobState.DOWNLOADED
                        set_output = {
                            "$set": {
                                "state": next_state.value,
                                "remote.process_id": process_running_uuid,
                            }
                        }
                        lock.update_on_release = set_output
                batch_manager.delete_terminated(
                    [(job_id, job_index, process_running_uuid)]
                )

    def ping_running_runner(self):
        ping_result = self.job_controller.ping_running_runner()
        if not ping_result:
            logger.info("Could not ping the running_runner document")

    def cleanup(self) -> None:
        """Close all the connections after stopping the Runner."""
        for worker_name, host in self.hosts.items():
            try:
                host.close()
            except Exception:
                logging.exception(
                    f"error while closing connection to worker {worker_name}"
                )

        try:
            self.jobstore.close()
        except Exception:
            logging.exception("error while closing connection to jobstore")

        self.job_controller.close()
